from __future__ import annotations

from dataclasses import asdict
from functools import partial
from pathlib import Path
from typing import Iterable, Optional, Union

import flax.linen as nn
import jax.numpy as jnp
import optax
import orbax.checkpoint
from flax.training import orbax_utils
from tqdm import tqdm

import lacss.metrics
from lacss.losses import (
    aux_size_loss,
    collaborator_border_loss,
    collaborator_segm_loss,
    lpn_loss,
    segmentation_loss,
)

from ..modules import Lacss, LacssCollaborator
from ..typing import Array, Optimizer
from .base_trainer import Trainer
from .strategy import JIT


class _CKSModel(nn.Module):
    principal: Lacss
    collaborator: LacssCollaborator

    def __call__(self, image, cls_id=None, gt_locations=None, *, training=False):
        outputs = self.principal(
            image,
            gt_locations=gt_locations,
            training=training,
        )

        if self.collaborator is not None:
            outputs.update(self.collaborator(image=image, cls_id=cls_id))
        return outputs


class LacssTrainer(Trainer):
    """Main trainer class for Lacss"""

    default_lr: float = 0.001

    def __init__(
        self,
        config: dict = {},
        collaborator_config: Optional[dict] = None,
        *,
        optimizer: Optional[Optimizer] = None,
        seed: Union[int, Array] = 42,
        strategy: type = JIT,
    ):
        """Constructor

        Args:
            config: configuration dictionary for Lacss model
            collaborator_config: configuration dictionary for the collaborator model used in
                weakly-supervised training. If set to None, then no collaborator model will be
                created. In this case, training with weak-supervision will result in a error.

        Keyword Args:
            seed: RNG seed
            optimizer: Override the default optimizer
            strategy: Training backend. See See [Traing backends](/api/train/#training-backends).
        """
        principal = Lacss.from_config(config)

        if collaborator_config is None:
            collaborator = None
        else:
            collaborator = LacssCollaborator(**collaborator_config)

        super().__init__(
            model=_CKSModel(
                principal=principal,
                collaborator=collaborator,
            ),
            optimizer=optimizer,
            seed=seed,
            strategy=strategy,
        )

        if self.optimizer is None:
            self.optimizer = optax.adamw(LacssTrainer.default_lr)

        # self._cp_step = 0

    def _train_to_next_interval(
        self, train_iter, checkpoint_interval, checkpoint_dir, val_dataset
    ):
        cur_step = self.state.step
        next_cp_step = (
            (cur_step + checkpoint_interval)
            // checkpoint_interval
            * checkpoint_interval
        )

        for _ in tqdm(range(cur_step, next_cp_step)):
            logs = next(train_iter)

        print(", ".join([f"{k}:{v:.4f}" for k, v in logs.items()]))

        if checkpoint_dir is not None:
            cps = [int(p.name.split("-")[-1]) for p in checkpoint_dir.glob("chkpt-*")]
            cp_cnt = max(cps) + 1
            self.checkpoint(checkpoint_dir / f"chkpt-{cp_cnt}")

        if val_dataset is not None:
            val_metrics = [
                lacss.metrics.LoiAP([5, 2, 1]),
                lacss.metrics.BoxAP([0.5, 0.75]),
            ]

            var_logs = self.test_and_compute(val_dataset, metrics=val_metrics)
            for k, v in var_logs.items():
                print(f"{k}: {v}")

    def _validate(self, val_dataset):
        if val_dataset is not None:
            val_metrics = [
                lacss.metrics.LoiAP([5, 2, 1]),
                lacss.metrics.BoxAP([0.5, 0.75]),
            ]

            var_logs = self.test_and_compute(
                val_dataset, metrics=val_metrics, strategy=JIT
            )
            for k, v in var_logs.items():
                print(f"{k}: {v}")

    def _make_checkpoint(self, checkpoint_manager):
        if checkpoint_manager is not None:
            lastest_step = checkpoint_manager.latest_step()
            if lastest_step is None:
                lastest_step = 0

            save_args = orbax_utils.save_args_from_target(self.state)
            checkpoint_manager.save(
                lastest_step + 1,
                {"config": asdict(self.model), "train_state": self.state},
                save_kwargs={"train_state": {"save_args": save_args}},
            )

    def restore_from_checkpoint(
        self, checkpoint_manager: orbax.checkpoint.CheckpointManager, step: int = -1
    ) -> None:
        """Restore train state from a checkpoint.

        Args:
            checkpoint_manager: Orbax checkpoint manager
            step: The exact checkpoint to restore. Latest if unspecified.
        """
        if step < 0:
            step = checkpoint_manager.latest_step()

        # restore_args = orbax_utils.restore_args_from_target(self.state, None)

        restored = checkpoint_manager.restore(
            step,
            items=dict(
                config={},
                train_state=self.state,
            ),
            # restore_kwargs={"train_state": {"restore_args": restore_args}},
        )
        self.state = restored["train_state"]

        # reconstruct model or not?
        # from lacss.utils import dataclass_from_dict
        # self.model = dataclass_from_dict(restored["config"])

        # self._cp_step = step

    def reset(
        self,
        warmup_steps: int = 0,
        sigma: float = 20.0,
        pi: float = 2.0,
    ) -> None:
        cur_step = self.state.step

        if cur_step >= warmup_steps:
            col_seg_loss = partial(collaborator_segm_loss, sigma=sigma, pi=pi)
            # col_seg_loss.name = "collaborator_segm_loss"
            self.losses = [
                lpn_loss,
                segmentation_loss,
                col_seg_loss,
                collaborator_border_loss,
                aux_size_loss,
            ]
        else:
            pre_seg_loss = partial(segmentation_loss, pretraining=True)
            pre_col_seg_loss = partial(collaborator_segm_loss, sigma=100.0, pi=1.0)
            self.losses = [
                lpn_loss,
                pre_seg_loss,
                pre_col_seg_loss,
                collaborator_border_loss,
                aux_size_loss,
            ]

        super().reset()

    def do_training(
        self,
        dataset: Iterable,
        val_dataset: Iterable | None = None,
        n_steps: int = 50000,
        validation_interval: int = 5000,
        checkpoint_manager: Optional[orbax.checkpoint.CheckpointManager] = None,
        *,
        warmup_steps: int = 0,
        sigma: float = 20.0,
        pi: float = 2.0,
    ) -> None:
        """Runing training.

        Args:
            dataset: An data iterator feed training data. The data should be in the form of a tuple:
                (x, y).
                    x is a dictionary with at least two keys:
                        "image": Trainging image.
                        "gt_locations": Point labels. Nx2 array
                    y is a dictionary of extra labels. It can be
                        None: for point-supervised training
                        "gt_labels": A index-label image (H, W). For segmentation label.
                        "gt_image_mask": A binary image (H, W). For weakly supervised training
            n_steps: Total training steps
            validation_inteval: Step intervals to perform validation and checkpointing.
            val_dataset: If not None, performing validation on this dataset. The data should be in the
                form of a tuple (x, y):
                    x is a dictionary with one key: "image"
                    y is a dictionary with two lables: "gt_bboxes" and "gt_locations"
            checkpoint_manager: If supplied, will be used to created checkpoints. A checkpoint manager
                can be obtained by calling:
                ```
                options = orbax.CheckpointManagerOptions(...)
                manager = orbax.checkpoint.CheckpointManager(
                    'path/to/directory/',
                    trainer.get_checkpointer(),
                    options = options
                )
                ```


        keyword Args:
            warmup_steps: Only used for point-supervised training. Pretraining steps, for which a large
                sigma values is used. This should be multiples of validation_inteval
            sigma: Only for point-supervised training. Expected cell size
            pi: Only for point-supervised training. Amplitude of the prior term.
        """
        train_iter = self.train(dataset, rng_cols=["droppath"], training=True)

        while self.state.step < n_steps:
            cur_step = self.state.step
            next_cp_step = (
                (cur_step + validation_interval)
                // validation_interval
                * validation_interval
            )

            self.reset(
                warmup_steps=warmup_steps,
                sigma=sigma,
                pi=pi,
            )

            print(f"Current step {cur_step} going to {next_cp_step}")
            for _ in tqdm(range(cur_step, next_cp_step)):
                logs = next(train_iter)

            print(", ".join([f"{k}:{v:.4f}" for k, v in logs.items()]))
            self._make_checkpoint(checkpoint_manager)
            self._validate(val_dataset)

    @classmethod
    def get_checkpointer(cls) -> orbax.checkpoint.Checkpointer:
        """Returns a checkpointer obj for this Trainer. Convienent function for creating a checkpoint manager"""

        return {
            "config": orbax.checkpoint.Checkpointer(
                orbax.checkpoint.JsonCheckpointHandler()
            ),
            "train_state": orbax.checkpoint.PyTreeCheckpointer(),
        }

    @classmethod
    def from_checkpoint(cls, cp_path) -> "LacssTrainer":
        """load the module and its weight from a checkpoint
            This utility static method allows use checkpoint as a model save. It ignores
            optstate.

        Args:
            cp_path: checkpoint location (a dir)

        Returns:
            LacssTrainer.
        """
        import json

        cp_path = Path(cp_path)

        with open(cp_path / "config" / "metadata", "r") as f:
            cfg = json.load(f)

        obj = cls(cfg["principal"], cfg["collaborator"])

        fake_data = dict(
            image=jnp.zeros([64, 64, 3]),
            gt_locations=jnp.zeros([8, 2]),
            cls_id=jnp.asarray(0),
        )
        obj.initialize([fake_data], tx=optax.adam(cls.default_lr))

        # restore_args = orbax_utils.restore_args_from_target(obj.state, None)
        obj.state = orbax.checkpoint.PyTreeCheckpointer().restore(
            cp_path / "train_state",
            item=obj.state,
            # restore_args = restore_args,
            # transforms={},
        )

        return obj

    def save(self, save_path) -> None:
        """Save a pickled copy of the Lacss model in the form of (module:Lacss, weights:FrozenDict). Only saves the principal model.

        Args:
            save_path: Path to the pkl file
        """
        import pickle

        _module = self.model.principal
        _params = self.params["principal"]

        with open(save_path, "wb") as f:
            pickle.dump((_module, _params), f)



from __future__ import annotations

from typing import Sequence

import jax
import jax.numpy as jnp

from ..typing import *

Shape = Sequence[int]


def locations_to_labels(
    locations: ArrayLike, target_shape: Shape, threshold: float = 1.5
) -> tuple[Array, Array]:
    """Generate labels as LPN regression targets
    
    Args:
        locations: [N, 2] float32 true location values. scaled 0..1, masking out invalid with -1
        target_shape: (H, W)  int
        threshold: distance threshold for postive label
    
    Returns:
        score_target: [H, W, 1] int32
        regression_target: [H, W, 2] float tensor
    """
    height, width = target_shape
    threshold_sq = threshold * threshold

    flat_locations = locations * jnp.array(target_shape)

    mesh = jnp.mgrid[:height, :width] + 0.5  # [2, H, W]
    distances = flat_locations[:, :, None, None] - mesh  # [N, 2, H, W]
    distances_sq = (distances * distances).sum(axis=1)  # [N, H, W]

    # masking off invalid
    distances_sq = jnp.where(
        flat_locations[:, None, None, 0] >= 0, distances_sq, float("inf")
    )

    indices = jnp.argmin(distances_sq, axis=0, keepdims=True)  # [1, H, W]
    best_distances = jnp.take_along_axis(distances_sq, indices, 0)
    score_target = (best_distances < threshold_sq).astype(int)  # [1, H, W]
    score_target = score_target[0][:, :, None]  # [H, W, 1]

    indices = indices.repeat(2, axis=0)[None, ...]  # [1, 2, H, W]
    regression_target = jnp.take_along_axis(distances, indices, 0).squeeze(
        0
    )  # [2, H, W]
    regression_target = regression_target.transpose(1, 2, 0)  # [H, W, 2]

    return score_target, regression_target


def distance_similarity(pred_locations: ArrayLike, gt_locations: ArrayLike) -> Array:
    """Compute distance similarity matrix

        pairwise similarity = 1 / distance ^2

    Args:
        pred_locations: [N, 2] use -1 to mask out invalid locations
        gt_locations: [K, 2] use -1 to mask out invalid locations

    Returns:
        similarity_matrix: [N, k]
    """

    distances_matrix = pred_locations[:, None, :] - gt_locations
    distances_matrix = jnp.square(distances_matrix).sum(axis=-1)

    sm = 1.0 / (distances_matrix + 1e-8)

    # negative locations are invalid
    sm = jnp.where((pred_locations < 0).all(axis=-1)[:, None], 0, sm)
    sm = jnp.where((gt_locations < 0).all(axis=-1), 0, sm)

    return sm


def location_matching(
    pred_locations: ArrayLike, gt_locations: ArrayLike, threshold: float
) -> tuple[Array, Array]:
    """Match predicted location to gt locations

    Args:
      pred_locations:r [N, 2]
      gt_locations: [K, 2]
      threshold: float. Maximum distance to be matched
    
    Returns:
      matches: [N], indices of the matches location in gt list
      indicators: [N] bool
    """

    threshold = 1.0 / (threshold * threshold)

    distances_matrix = distance_similarity(pred_locations, gt_locations)

    matches = distances_matrix.argmax(axis=-1, keepdims=True)
    indicators = jnp.take_along_axis(distances_matrix, matches, -1) > threshold

    return matches.squeeze(-1), indicators.squeeze(-1)



import numpy as np
import pandas as pd


def decode_patch(codes, h, w):
    # decode string to a array representing the mask
    # inputs:
    #   codes: string
    #   h: height of patch, int
    #   w: width of patch, int
    # outputs:
    #   image:  array of hxw
    return (np.array([*codes]) == "1").astype(int).reshape(h, w)


def extract_frame(df, frame_no, img_height=512, img_width=512):
    # extract a single frame, label all cells
    # inputs:
    #   df: dataframe
    #   frame_no: int
    #   image_height: int
    #   image_width: int
    #  outputs:
    #   image: 2d array of [image_height, image_width]
    df_1 = df.loc[df["frame"] == frame_no]
    y0s = np.array(df_1["bbox_y"]).astype(int)
    x0s = np.array(df_1["bbox_x"]).astype(int)
    heights = np.array(df_1["bbox_h"]).astype(int)
    widths = np.array(df_1["bbox_w"]).astype(int)
    patches = np.array(df_1["image"])

    image = np.zeros([img_height, img_width], dtype=int)
    for k, (p, h, w, y, x) in enumerate(zip(patches, heights, widths, y0s, x0s)):
        decoded = decode_patch(p, h, w)
        decoded = decoded * (k + 1)
        image[y : y + h, x : x + w] = np.maximum(decoded, image[y : y + h, x : x + w])
    return image


def _data_generator(imgs, model, nms):

    from tqdm import tqdm

    from ..ops import non_max_suppress_predictions
    from ..train.strategy import JIT
    from ..utils import format_predictions

    for frame, img in enumerate(tqdm(imgs)):
        inputs = dict(image=img)
        pred = model(inputs)
        if nms:
            mask = non_max_suppress_predictions(pred)
        else:
            mask = None

        results = format_predictions(pred, mask)

        for k in range(len(results["locations"])):
            yield (
                frame + 1,
                results["locations"][k, 0],
                results["locations"][k, 1],
                results["scores"][k],
                # results['centroids'][k, 0],
                # results['centroids'][k, 1],
                results["bboxes"][k, 0],
                results["bboxes"][k, 1],
                results["bboxes"][k, 2] - results["bboxes"][k, 0],
                results["bboxes"][k, 3] - results["bboxes"][k, 1],
                results["encodings"][k].count("1"),
                results["encodings"][k],
            )


def generate_predictions(imgs, model, nms=False):
    """segmentation predictions of a movie
    Args:
        imgs: an iterator of images sorted by frames
        model: a segmentation predictor or trainer
        nms: bool whether to perform nms
    Returns:
        pandas dataframe
    """
    df = pd.DataFrame(
        _data_generator(imgs, model, nms),
        columns=[
            "frame",
            "y",
            "x",
            "score",
            "bbox_y",
            "bbox_x",
            "bbox_h",
            "bbox_w",
            "area",
            "image",
        ],
    )

    return df


def to_labels(df, image_heigh=512, image_width=512):
    """
    Args:
        df: segmentation predictions
        image_height: int
        image_wdith: int
    returns:
        an iterator of image labels
    """

    max_frame = int(df["frame"].max())
    for frame in range(1, max_frame + 1):
        label = extract_frame(
            df,
            frame,
        )
        yield label



from .base_trainer import Trainer
from .lacss_mt_trainer import LacssMTTrainer
from .lacss_trainer import LacssTrainer
from .strategy import JIT, Core, Distributed, Eager, VMapped
from .utils import Inputs, TFDatasetAdapter, TorchDataLoaderAdapter



from __future__ import annotations

import glob
import json
from functools import partial
from os.path import join
from pathlib import Path
from typing import Iterator, Optional, Sequence

import imageio.v2 as imageio
import numpy as np
import tensorflow as tf
from skimage.measure import regionprops


# A adaptation of tf.image.crop_and_resize to use edge-based indexing
# Edge-based indexing is more accurate for float-value bboxes
# i.e. value (0, 0) refers to the top-left cornor (not center) of the top-left pixel
def _crop_and_resize(masks, boxes, target_shape):
    box_h = boxes[:, 2] - boxes[:, 0]
    box_w = boxes[:, 3] - boxes[:, 1]
    dh = box_h / target_shape[0] / 2
    dw = box_w / target_shape[1] / 2
    boxes = tf.stack([dh, dw, -dh, -dw], axis=-1) + boxes - 0.5
    H = tf.shape(masks)[1]
    W = tf.shape(masks)[2]
    segs = tf.image.crop_and_resize(
        masks[..., None],
        boxes / [H - 1, W - 1, H - 1, W - 1],
        tf.range(tf.shape(boxes)[0]),
        target_shape,
    )
    return tf.squeeze(segs, axis=-1)


def coco_generator_full(
    annotation_file: str,
    image_path: str,
    mask_shape: Optional[tuple[int, int]] = None,
) -> Iterator[dict]:
    """A generator function to produce coco-annotated data

    Args:
        annotation_file: Path to coco annotation files
        image_path: Path to image directory
        mask_shape: If supplied, all the instance segmentations will be croped and resized to the specifed size. Otherwise,
            the segmentations are uncropped (in original image size)

    Yields:
        A data dictionary with thse keys:

            - id: data id
            - filename: image filename
            - image: an array [H, W, C]
            - masks: segmentation masks. [N, H, W] or [N,] + mask_shape
            - centroids: yx format.
            - bboxes: y0x0y1x1 format.
            - label: an array [H, W] representing pixel labels of all instances.
    """
    from pycocotools.coco import COCO

    coco = COCO(annotation_file=annotation_file)
    for imgid in coco.getImgIds():
        imginfo = coco.imgs[imgid]
        bboxes = []
        locs = []
        masks = []
        # segs = []
        for ann_id in coco.getAnnIds(imgIds=imgid):
            ann = coco.anns[ann_id]
            bbox = ann["bbox"]
            bbox = np.array([bbox[1], bbox[0], bbox[1] + bbox[3], bbox[0] + bbox[2]])
            bboxes.append(bbox)

            mask = coco.annToMask(ann)
            masks.append(mask)

            mi = np.stack(np.where(mask >= 0.5), axis=-1)
            locs.append(mi.mean(axis=0) + 0.5)

        bboxes = np.array(bboxes, dtype="float32")
        locs = np.array(locs, dtype="float32")
        masks = np.array(masks, dtype="float32")

        filepath = glob.glob(join(image_path, "**", imginfo["file_name"]))
        assert len(filepath) == 1
        image = imageio.imread(filepath[0])

        if len(image.shape) == 2:
            image = image[:, :, None]
        image = (image / 255.0).astype("float32")

        img_h, img_w, _ = image.shape
        bboxes = np.clip(
            bboxes / [img_h, img_w, img_h, img_w],
            0.0,
            1.0,
        ) * [img_h, img_w, img_h, img_w]
        bboxes = bboxes.astype("float32")

        locs = np.clip(
            locs / [img_h, img_w],
            0.0,
            1.0,
        ) * [img_h, img_w]
        locs = locs.astype("float32")

        if mask_shape is not None:
            segs = _crop_and_resize(masks, bboxes, mask_shape).numpy()
        else:
            segs = masks

        n_cells = masks.shape[0]
        masks = masks.astype("int32") * np.arange(1, n_cells + 1).reshape(n_cells, 1, 1)
        label = masks.max(axis=0).astype("int32")

        yield {
            "id": imgid,
            "filename": imginfo["file_name"],
            "image": image,
            "masks": segs,
            "centroids": locs,
            "bboxes": bboxes,
            "label": label,
        }


def coco_generator(annotation_file: str, image_path: str) -> Iterator[dict]:
    from pycocotools.coco import COCO

    coco = COCO(annotation_file=annotation_file)
    for imgid in coco.getImgIds():
        imginfo = coco.imgs[imgid]
        bboxes = []
        segs = []
        for ann_id in coco.getAnnIds(imgIds=imgid):
            ann = coco.anns[ann_id]
            bbox = ann["bbox"]
            bbox = [bbox[1], bbox[0], bbox[1] + bbox[3], bbox[0] + bbox[2]]
            bboxes.append(bbox)

            seg = np.reshape(coco.anns[ann_id]["segmentation"], [-1, 2])
            segs.append(seg)

        filepath = glob.glob(join(image_path, "**", imginfo["file_name"]))
        assert len(filepath) == 1

        image = imageio.imread(filepath[0])

        if len(image.shape) == 2:
            image = image[:, :, None]
        image = (image / 255.0).astype("float32")

        yield {
            "id": imgid,
            "filename": imginfo["file_name"],
            "image": image,
            "bboxes": bboxes,
            "polygons": segs,
        }


def dataset_from_coco_annotations(
    annotation_file: str,
    image_path: str,
    image_shape: tuple = [None, None, 3],
    mask_shape: tuple = [48, 48],
) -> tf.Dataset:
    """Obtaining a tensowflow dataset from coco annotations. See [coco_generator_full()](./#lacss.data.generator.coco_generator_full)

    Args:
        annotation_file: Path to coco annotation files
        image_path: Path to image directory
        image_shape: The expect image shapes. Use None to represent variable dimensions.
        mask_shape: If supplied, all the instance segmentations will be croped and resized to the specifed size. Otherwise,
            the segmentations are uncropped (in original image size)

    Returns:
        A tensorflow dataset.

    """
    if mask_shape is None:
        mask_spec = tf.TensorSpec([None] + image_shape[:2], dtype=tf.float32)
    else:
        mask_spec = tf.TensorSpec([None] + mask_shape, dtype=tf.float32)

    return tf.data.Dataset.from_generator(
        lambda: coco_generator_full(annotation_file, image_path, mask_shape=mask_shape),
        output_signature={
            "id": tf.TensorSpec([], dtype=tf.int64),
            "filename": tf.TensorSpec([], dtype=tf.string),
            "image": tf.TensorSpec(image_shape, dtype=tf.float32),
            "masks": mask_spec,
            "centroids": tf.TensorSpec([None, 2], dtype=tf.float32),
            "bboxes": tf.TensorSpec([None, 4], dtype=tf.float32),
            "label": tf.TensorSpec(image_shape[:2], dtype=tf.int32),
        },
    )


def simple_generator(annotation_file: str, image_path: str) -> Iterator[dict]:
    """A simple generator function to produce image data labeled with points and image-level segmentaion.

    Args:
        annotation_file: Path to the json format annotation file.
        image_path: Path to the image directory.

    Yields:
        A data dictionary with thse keys:

            * img_id: data id
            * image: an array [H, W, C]
            * centroids: yx format.
            * image_mask: segmentation masks for the image. [H, W]
    """

    with open(annotation_file, "r") as f:
        annotations = json.load(f)

    for k, ann in enumerate(annotations):
        image = imageio.imread(join(image_path, ann["image_file"]))
        if len(image.shape) == 2:
            image = image[:, :, None]
        image = (image / 255.0).astype("float32")

        if "mask_file" in ann:
            binary_mask = imageio.imread(join(image_path, ann["mask_file"]))

            if len(binary_mask.shape) == 3:
                binary_mask = binary_mask[:, :, 0]
            binary_mask = (binary_mask > 0).astype("float32")
        else:
            binary_mask = np.zeros(image.shape[:2], dtype="float32")

        locations = np.array(ann["locations"]).astype("float32")

        if "img_id" in ann:
            img_id = ann["img_id"]
        else:
            img_id = k

        yield {
            "img_id": img_id,
            "image": image,
            "centroids": locations,
            "image_mask": binary_mask,
        }


def dataset_from_simple_annotations(
    annotation_file: str,
    image_path: str,
    image_shape=[None, None, 3],
) -> tf.Dataset:
    """Obtaining a tensowflow dataset from simple annotatiion. See [simple_generator()](./#lacss.data.generator.simple_generator)

    Args:
        annotation_file: Path to the json format annotation file.
        image_path: Path to the image directory.
        image_shape: The expect image shapes. Use None to represent variable dimensions.

    Returns:
        A tensorflow dataset object

    """

    output_signature = {
        "img_id": tf.TensorSpec([], dtype=tf.int64),
        "image": tf.TensorSpec(image_shape, dtype=tf.float32),
        "centroids": tf.TensorSpec([None, 2], dtype=tf.float32),
        "image_mask": tf.TensorSpec(image_shape[:2], dtype=tf.float32),
    }

    return tf.data.Dataset.from_generator(
        lambda: simple_generator(annotation_file, image_path),
        output_signature=output_signature,
    )


def img_mask_pair_generator(
    imgfiles: Sequence[str | Path],
    maskfiles: Sequence[str | Path],
) -> Iterator[dict]:
    """A generator function to produce image data labeled with segmentation labels.
        In this case, one has paired input images and label images as files on disk.

    Args:
        imgfiles: List of file pathes to input image file.
        maskfiles: List of file pathes to label image file.

    Yields:
        A data dictionary with thse keys:

            * img_id: data id
            * image: an array [H, W, C]
            * centroids: yx format.
            * bboxes: y0x0y1x1 format.
            * label: an array [H, W] representing pixel labels of all instances.
    """
    for k, (img_file, mask_file) in enumerate(zip(imgfiles, maskfiles)):
        img = imageio.imread(img_file).astype("float32")
        img /= 255
        if len(img.shape) == 2:
            img = img[:, :, None]

        mask = imageio.imread(mask_file)
        if len(mask.shape) == 3:
            mask = mask[:, :, 0]

        bboxes = []
        locs = []
        for prop in regionprops(mask):
            bboxes.append(prop["bbox"])
            locs.append(prop["centroid"])

        yield {
            "img_id": k,
            "image": img.astype("float32"),
            "centroids": np.asarray(locs, dtype="float32") + 0.5,
            "bboxes": np.asarray(bboxes),
            "label": mask,
        }


# def _bbox_from_label(label):
#     import numpy as np
#     from skimage.measure import regionprops

#     bboxes = np.asarray([r["bbox"] for r in regionprops(label)])
#     centroids = np.asarray([r["centroids"] for r in regionprops(label)])

#     return bboxes, centroids


def _mask_from_label(inputs, *, mask_shape=[48, 48]):
    label = inputs["label"]

    # if not "bboxes" in inputs or not "centroids" in inputs:
    #     inputs["bboxes"], inputs["centroids"] = tf.numpy_func(
    #         _bbox_from_label,
    #         [label],
    #         [tf.float32, tf.float32],
    #         False,
    #     )

    n_instances = tf.shape(inputs["bboxes"])[0]

    label_expanded = tf.cast(
        label == tf.range(1, n_instances + 1)[:, None, None],
        tf.int32,
    )

    inputs["masks"] = _crop_and_resize(
        label_expanded,
        inputs["bboxes"],
        mask_shape,
    )

    return inputs


def dataset_from_img_mask_pairs(
    imgfiles: Sequence[str | Path],
    maskfiles: Sequence[str | Path],
    *,
    image_shape: Sequence[int | None] = [None, None, 3],
    generate_masks: bool = False,
    mask_shape: tuple[int, int] = [48, 48],
) -> tf.Dataset:
    """Obtaining a tensowflow dataset from image/label pairs.
            See [img_mask_pair_generator()](./#lacss.data.generator.img_mask_pair_generator)

    Args:
        imgfiles: List of file pathes to input image file.
        maskfiles: List of file pathes to label image file.

    Keyword Args:
        image_shape: The expect image shapes. Use None to represent variable dimensions.
        generate_masks: Whether to convert label images to indiviudal instance masks. This
            should be True if your data augmentation pipeline include rescaling ops or cropping
            ops., becasue these ops does not recompute the label image.
        mask_shape: Only when generate_mask=True. The resolution of the generated masks.

    Returns:
        A tensorflow dataset object
    """
    ds = tf.data.Dataset.from_generator(
        lambda: img_mask_pair_generator(imgfiles, maskfiles),
        output_signature={
            "img_id": tf.TensorSpec([], dtype=tf.int64),
            "image": tf.TensorSpec(image_shape, dtype=tf.float32),
            "centroids": tf.TensorSpec([None, 2], dtype=tf.float32),
            "bboxes": tf.TensorSpec([None, 4], dtype=tf.float32),
            "label": tf.TensorSpec(image_shape[:2], dtype=tf.int32),
        },
    )

    if generate_masks:
        ds = ds.map(lambda x: _mask_from_label(x, mask_shape=mask_shape))

    return ds



from __future__ import annotations

from collections import defaultdict
from dataclasses import asdict, field

import flax.linen as nn
import jax.numpy as jnp

from ..ops import *
from ..typing import *
from .convnext import ConvNeXt
from .detector import Detector
from .lpn import LPN
from .segmentor import Segmentor


class Lacss(nn.Module):
    """Main class for LACSS model

    Attributes:
        backbone: The ConvNeXt backbone
        lpn: The LPN head for detecting cell location
        detector: A weight-less module interpreting lpn output
        segmentor: The segmentation head

    """

    backbone: ConvNeXt = field(default_factory=ConvNeXt)
    lpn: LPN = field(default_factory=LPN)
    detector: Detector = field(default_factory=Detector)
    segmentor: Segmentor = field(default_factory=Segmentor)

    @classmethod
    def from_config(cls, config: dict):
        """Factory method to build an Lacss instance from a configuration dictionary

        Args:
            config: A configuration dictionary.

        Returns:
            An Lacss instance.

        """
        config_ = defaultdict(lambda: {})
        config_.update(config)
        return Lacss(
            backbone=ConvNeXt(**config_["backbone"]),
            lpn=LPN(**config_["lpn"]),
            detector=Detector(**config_["detector"]),
            segmentor=Segmentor(**config_["segmentor"]),
        )
        # return dataclass_from_dict(cls, config)

    def get_config(self) -> dict:
        """Convert to a configuration dict. Can be serialized with json

        Returns:
            config: a configuration dict
        """
        return asdict(self)

    def __call__(
        self,
        image: ArrayLike,
        gt_locations: ArrayLike | None = None,
        *,
        training: bool | None = None,
    ) -> dict:
        """
        Args:
            image: [H, W, C]
            gt_locations: [M, 2] if training, otherwise None
        Returns:
            a dict of model outputs

        """
        orig_height, orig_width, ch = image.shape
        if ch == 1:
            image = jnp.repeat(image, 3, axis=-1)
        elif ch == 2:
            image = jnp.concatenate([image, jnp.zeros_like(image[..., :1])], axis=-1)

        # assert image.shape[-1] == 3

        # ensure input size is multiple of 32
        height = ((orig_height - 1) // 32 + 1) * 32
        width = ((orig_width - 1) // 32 + 1) * 32
        image = jnp.pad(
            image, [[0, height - orig_height], [0, width - orig_width], [0, 0]]
        )

        # backbone
        encoder_features, features = self.backbone(image, training=training)
        model_output = dict(
            encoder_features=encoder_features,
            decoder_features=features,
        )

        # detection
        scaled_gt_locations = (
            gt_locations / jnp.array([height, width])
            if gt_locations is not None
            else None
        )
        model_output.update(
            self.lpn(
                inputs=features,
                scaled_gt_locations=scaled_gt_locations,
            )
        )

        if gt_locations is not None or not training:

            model_output.update(
                self.detector(
                    scores=model_output["lpn_scores"],
                    regressions=model_output["lpn_regressions"],
                    gt_locations=gt_locations,
                    training=training,
                )
            )

            # segmentation
            locations = model_output[
                "training_locations" if gt_locations is not None else "pred_locations"
            ]
            scaled_locs = locations / jnp.array([height, width])
            model_output.update(
                self.segmentor(
                    features=features,
                    locations=scaled_locs,
                )
            )

        return model_output



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: lacss.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0blacss.proto\x12\x0ftrackmate.lacss\"l\n\x05Image\x12\x0e\n\x06height\x18\x01 \x01(\x04\x12\r\n\x05width\x18\x02 \x01(\x04\x12\x0f\n\x07\x63hannel\x18\x03 \x01(\x04\x12%\n\x05\x64type\x18\x04 \x01(\x0e\x32\x16.trackmate.lacss.DType\x12\x0c\n\x04\x64\x61ta\x18\x05 \x01(\x0c\"\xb5\x01\n\x08Settings\x12\x15\n\rmin_cell_area\x18\x01 \x01(\x02\x12\x1b\n\x13remove_out_of_bound\x18\x02 \x01(\x08\x12\x0f\n\x07scaling\x18\x03 \x01(\x02\x12\x0f\n\x07nms_iou\x18\x04 \x01(\x02\x12\x1b\n\x13\x64\x65tection_threshold\x18\x05 \x01(\x02\x12\x1e\n\x16segmentation_threshold\x18\x06 \x01(\x02\x12\x16\n\x0ereturn_polygon\x18\x07 \x01(\x08\"[\n\x05Input\x12+\n\x08settings\x18\x01 \x01(\x0b\x32\x19.trackmate.lacss.Settings\x12%\n\x05image\x18\x02 \x01(\x0b\x32\x16.trackmate.lacss.Image\"4\n\x05Label\x12\x0e\n\x06height\x18\x01 \x01(\x04\x12\r\n\x05width\x18\x02 \x01(\x04\x12\x0c\n\x04\x64\x61ta\x18\x06 \x01(\x0c\"V\n\x06Result\x12%\n\x05score\x18\x01 \x01(\x0b\x32\x16.trackmate.lacss.Label\x12%\n\x05label\x18\x02 \x01(\x0b\x32\x16.trackmate.lacss.Label\"\x1d\n\x05Point\x12\t\n\x01x\x18\x01 \x01(\x02\x12\t\n\x01y\x18\x02 \x01(\x02\"@\n\x07Polygon\x12\r\n\x05score\x18\x01 \x01(\x02\x12&\n\x06points\x18\x02 \x03(\x0b\x32\x16.trackmate.lacss.Point\";\n\rPolygonResult\x12*\n\x08polygons\x18\x01 \x03(\x0b\x32\x18.trackmate.lacss.Polygon*\x14\n\x05\x44Type\x12\x0b\n\x07\x46LOAT32\x10\x00\x42\x30\n$fiji.plugin.trackmate.detector.lacssB\x08LacssMsgb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'lacss_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  _globals['DESCRIPTOR']._options = None
  _globals['DESCRIPTOR']._serialized_options = b'\n$fiji.plugin.trackmate.detector.lacssB\010LacssMsg'
  _globals['_DTYPE']._serialized_start=719
  _globals['_DTYPE']._serialized_end=739
  _globals['_IMAGE']._serialized_start=32
  _globals['_IMAGE']._serialized_end=140
  _globals['_SETTINGS']._serialized_start=143
  _globals['_SETTINGS']._serialized_end=324
  _globals['_INPUT']._serialized_start=326
  _globals['_INPUT']._serialized_end=417
  _globals['_LABEL']._serialized_start=419
  _globals['_LABEL']._serialized_end=471
  _globals['_RESULT']._serialized_start=473
  _globals['_RESULT']._serialized_end=559
  _globals['_POINT']._serialized_start=561
  _globals['_POINT']._serialized_end=590
  _globals['_POLYGON']._serialized_start=592
  _globals['_POLYGON']._serialized_end=656
  _globals['_POLYGONRESULT']._serialized_start=658
  _globals['_POLYGONRESULT']._serialized_end=717
# @@protoc_insertion_point(module_scope)



from __future__ import annotations

from typing import Iterable, Optional, Union

import jax
import jax.numpy as jnp
import numpy as np
import orbax
from tqdm import tqdm

from ..deploy import Predictor
from ..typing import Array, Optimizer
from ..utils import pack_x_y_sample_weight, unpack_x_y_sample_weight
from .lacss_trainer import LacssTrainer


def _get_iterator(g):
    try:
        it = iter(g)
    except:
        it = iter(g())

    return it


class LacssMTTrainer(LacssTrainer):
    def __init__(
        self,
        config: dict = {},
        collaborator_config: Optional[dict] = None,
        *,
        optimizer: Optional[Optimizer] = None,
        seed: Union[int, Array] = 42,
    ):
        super().__init__(config, collaborator_config, optimizer=optimizer, seed=seed)

    def _merged_ds(self, labeled_dataset, unlabeled_dataset):
        labeled_dataset = _get_iterator(labeled_dataset)
        unlabeled_dataset = _get_iterator(unlabeled_dataset)

        while True:
            yield next(labeled_dataset)

            x, y, sw = unpack_x_y_sample_weight(next(unlabeled_dataset))

            if isinstance(x, dict):
                x = x["image"]

            pred = self.predictor.predict(x)
            locs = np.asarray(pred["pred_locations"])
            scores = np.asarray(pred["pred_scores"])
            locs = np.where(scores[:, None] >= 0.5, locs, -1)

            x = dict(
                image=x,
                gt_locations=locs,
            )
            yield pack_x_y_sample_weight(x, y, sw)

    def do_training(
        self,
        datasets: tuple[Iterable, Iterable],
        val_dataset: Iterable = None,
        n_steps: int = 50000,
        validation_interval: int = 5000,
        checkpoint_manager: Optional[orbax.checkpoint.CheckpointManager] = None,
        *,
        warmup_steps: int = 0,
        sigma: float = 20.0,
        pi: float = 2.0,
        alpha: float = 0.999,
    ) -> None:
        pred_model = self.model.principal
        pred_params = self.params["principal"]
        self.predictor = Predictor((pred_model, pred_params))

        combined_dataset = self._merged_ds(*datasets)
        train_iter = self.train(combined_dataset, rng_cols=["droppath"], training=True)

        while self.state.step < n_steps:
            cur_step = self.state.step
            next_cp_step = (
                (cur_step + validation_interval)
                // validation_interval
                * validation_interval
            )

            self.reset(
                warmup_steps=warmup_steps,
                sigma=sigma,
                pi=pi,
            )

            @jax.jit
            def mt_merge(tree_a, tree_b, alpha):
                return jax.tree_util.tree_map(
                    lambda a, b: a * alpha + b * (1 - alpha),
                    tree_a,
                    tree_b,
                )

            print(f"Current step {cur_step} going to {next_cp_step}")
            for _ in tqdm(range(cur_step, next_cp_step)):
                logs = next(train_iter)

                self.predictor.params = mt_merge(
                    self.predictor.params,
                    self.params["principal"],
                    alpha,
                )

            print(", ".join([f"{k}:{v:.4f}" for k, v in logs.items()]))
            self._make_checkpoint(checkpoint_manager)
            self._validate(val_dataset)



""" Various functions deals with segmentation pathces

    All functions here takes unbatched input. Use vmap to convert to batched data
"""
from __future__ import annotations

from functools import partial
from typing import Optional, Sequence

import jax
import jax.numpy as jnp
import numpy as np

from ..typing import *
from .boxes import box_iou_similarity
from .image import sub_pixel_samples

Shape = Sequence[int]


def gather_patches(
    features: ArrayLike, locations: ArrayLike, patch_size: int
) -> tuple[Array, Array, Array, Array]:
    # """extract feature patches according to a list of locations

    # Args:
    #     features: [H,W,C] standard 2D feature map
    #     locations: [N, 2] float32, scaled 0..1
    #     patch_size: int

    # Returns:
    #     patches: [N, patch_size, patch_size, C]
    #     y0: [N]: y0 coordinates of patches
    #     x0: [N]: x0 coordinates of patches
    # """
    height, width, _ = features.shape

    locations *= jnp.array([height, width])
    # i_locations = (locations + .5).astype(int)
    i_locations = locations.astype(int)
    i_locations_x = jnp.clip(i_locations[:, 1], 0, width - 1)
    i_locations_y = jnp.clip(i_locations[:, 0], 0, height - 1)
    yy, xx = jnp.mgrid[:patch_size, :patch_size] - patch_size // 2
    xx += i_locations_x[:, None, None]
    yy += i_locations_y[:, None, None]
    remainder = locations - jnp.stack([i_locations_y, i_locations_x], axis=-1)

    # padding to avoid out-of-bound
    padding_size = patch_size // 2 + 1
    paddings = [
        [padding_size, padding_size],
        [padding_size, padding_size],
        [0, 0],
    ]
    padded_features = jnp.pad(features, paddings)

    patches = padded_features[yy + padding_size, xx + padding_size, :]

    return patches, yy[:, 0, 0], xx[:, 0, 0], remainder


def _get_patch_data(pred):
    if isinstance(pred, dict):
        patches = pred["instance_output"]
        yc = pred["instance_yc"]
        xc = pred["instance_xc"]
    else:
        patches, yc, xc = pred

    if patches.ndim > yc.ndim:
        patches = patches.squeeze(-1)

    return patches, yc, xc


def bboxes_of_patches(pred: Sequence | DataDict, threshold: float = 0.5) -> jnp.ndarray:
    """Compute the instance bboxes from model predictions

    Args:
        pred: A model prediction dictionary:
        threshold: for segmentation, default 0.5

    Returns:
        bboxes: [n, 4] bboox for empty patches are filled with -1
    """
    patches, yy, xx = _get_patch_data(pred)

    _, d0, d1 = patches.shape
    row_mask = jnp.any(patches > threshold, axis=1)
    col_mask = jnp.any(patches > threshold, axis=2)

    min_col = row_mask.argmax(axis=1)
    max_col = d1 - row_mask[:, ::-1].argmax(axis=1)

    min_row = col_mask.argmax(axis=1)
    max_row = d0 - col_mask[:, ::-1].argmax(axis=1)

    min_row += yy[:, 0, 0]
    max_row += yy[:, 0, 0]
    min_col += xx[:, 0, 0]
    max_col += xx[:, 0, 0]

    is_valid = row_mask.any(axis=1, keepdims=True)
    bboxes = jnp.stack([min_row, min_col, max_row, max_col], axis=-1)
    bboxes = jnp.where(is_valid, bboxes, -1)

    return bboxes


# def indices_of_patches(
#     pred: Sequence | DataDict,
#     input_size: tuple[int, int] = None,
#     threshold: float = 0.5,
# ) -> Array:
#     """Compute yx coodinates of all segmented instances

#     Args:
#         pred: A model prediction dictionary
#         threshold: float

#     Returns:
#         [N, 3] array. The first two columns are y-x coordinates. The third
#             column is a index value for different instances.
#     """
#     patches, yy, xx = _get_patch_data(pred)

#     indices = patches >= threshold
#     yy = yy[indices]
#     xx = xx[indices]
#     rowids = jnp.argwhere(indices)[:, 0]

#     if input_size is not None:
#         height, width = input_size
#         valid_coords = ((yy >= 0) & (yy < height)) & ((xx >= 0) & (xx < width))
#         yy = yy[valid_coords]
#         xx = xx[valid_coords]
#         rowids = rowids[valid_coords]

#     return jnp.stack([yy, xx, rowids], axis=-1)


# def iou_patches_and_labels(
#     pred: DataDict, labels: ArrayLike, BLOCK_SIZE: int = 128, threshold: float = 0.5
# ) -> Array:
#     """Compute iou between prediction and ground truth label.

#     Args:
#         pred: A model prediction dictionary
#         labels: image label. bg_label = 0
#         threshold: for segmentation. default is 0.5

#     Returns:
#         [n, m] array of iou values.
#     """
#     patches, yc, xc = _get_patch_data(pred)
#     patches = patches >= threshold
#     pred_areas = jnp.count_nonzero(patches, axis=(-1, -2))
#     labels = labels.astype(int)

#     pads = yc.shape[-1] // 2
#     padded_labels = jnp.pad(labels, pads, constant_values=0)
#     gt_patches = padded_labels[yc + pads, xc + pads]

#     max_indices = ((labels.max() - 1) // BLOCK_SIZE + 1) * BLOCK_SIZE
#     all_gt_areas = jnp.count_nonzero(
#         labels[:, :, None] == jnp.arange(max_indices) + 1, axis=(0, 1)
#     )

#     ious = []
#     # FIXME change to scan
#     for k in range(1, labels.max() + 1, BLOCK_SIZE):
#         gt_p = gt_patches == jnp.arange(k, k + BLOCK_SIZE).reshape(
#             -1, 1, 1, 1
#         )  # [B, N, s, s]
#         gt_areas = all_gt_areas[k - 1 : k + BLOCK_SIZE - 1]
#         intersect = jnp.count_nonzero(gt_p & patches, axis=(-1, -2))  # [B, N]
#         ious.append(intersect / (pred_areas + gt_areas[:, None] - intersect + 1.0e-8))

#     ious = jnp.concatenate(ious, axis=0).transpose()  # [N, B * b]
#     ious = ious[:, : labels.max()]  # this breaks JIT

#     return ious


def patches_to_segmentations(
    pred: DataDict, input_size: Shape, threshold: float = 0.5
) -> Array:
    """Expand the predicted patches to the full image size.
    The default model segmentation output shows only a small patch around each instance. This
    function expand each patch to the size of the orginal image.

    Args:
        pred: A model prediction dictionary
        input_size: shape of the input image. Tuple of H, W
        threshold: for segmentation. Default is 0.5.

    Returns:
        segmentations: [n, height, width] n full0-size segmenatation masks.

    """
    patches, yc, xc = _get_patch_data(pred)
    n_patches, patch_size, _ = yc.shape

    page_nums = jnp.arange(n_patches)
    segms = jnp.zeros((n_patches,) + input_size)
    segms = segms.at[page_nums[:, None, None], yc, xc].set(patches)

    return (segms >= threshold).astype(int)


def patches_to_label(
    pred: DataDict,
    input_size: Shape,
    mask: Optional[ArrayLike] = None,
    score_threshold: float = 0.5,
    threshold: float = 0.5,
) -> Array:
    """convert patch output to the image label

    Args:
        pred: A model prediction dictionary
        input_size: shape of the input image. Tuple of H, W
        mask: boolean indicators masking out unwanted instances. Default is None (all cells)
        score_threshold: otional, min_score to be included. Default is .5.
        threshold: segmentation threshold.  Default .5

    Returns:
        label: [height, width]
    """
    label = jnp.zeros(input_size)
    pr = pred["instance_output"] > threshold

    if mask is None:
        mask = pred["instance_mask"].squeeze(axis=(1, 2))
    else:
        mask &= pred["instance_mask"].squeeze(axis=(1, 2))

    if score_threshold > 0:
        mask &= pred["pred_scores"] >= score_threshold

    idx = jnp.cumsum(mask) * mask
    idx = jnp.where(mask, idx.max() + 1 - idx, 0)
    
    pr = (pr > threshold).astype(int) * idx[:, None, None]

    yc = pred["instance_yc"]
    xc = pred["instance_xc"]

    label = label.at[yc, xc].max(pr)
    label = jnp.where(label, label.max() + 1 - label, 0)

    return label


# def ious_of_patches_from_same_image(pred: DataDict, threshold: float = 0.5) -> Array:
#     """Compute IOUs among instances from the same image. Most likely used for nms.

#     Args:
#         pred: A model prediction dictionary
#         threshold: Segmentation threshold. Default is 0.5.

#     Returns:
#         ious: IOUs as an upper triagle matrix.
#     """

#     bboxes = bboxes_of_patches(pred)
#     box_ious = box_iou_similarity(bboxes, bboxes)
#     box_ious = np.triu(box_ious, 1)
#     # n_detections = jnp.count_nonzero(pred['instance_mask'])
#     # box_ious = box_ious[:n_detections, :n_detections]

#     cy, cx = np.where(box_ious > 0)
#     pad_size = pred["instance_output"].shape[-1]

#     patches = np.asarray(pred["instance_output"]) >= threshold
#     patches_y = patches[cy]
#     patches_x = patches[cx]
#     plt_to = np.zeros([len(cy), pad_size * 3, pad_size * 3], dtype=bool)
#     dy = (
#         np.asarray(pred["instance_yc"])[cy, 0, 0]
#         - np.asarray(pred["instance_yc"])[cx, 0, 0]
#         + pad_size
#     )
#     dx = (
#         np.asarray(pred["instance_xc"])[cy, 0, 0]
#         - np.asarray(pred["instance_xc"])[cx, 0, 0]
#         + pad_size
#     )

#     for k in range(len(cy)):
#         plt_to[k, dy[k] : dy[k] + pad_size, dx[k] : dx[k] + pad_size] = patches_y[k]
#     plt_to = plt_to[:, pad_size : pad_size * 2, pad_size : pad_size * 2]
#     unions = np.count_nonzero(plt_to & patches_x, axis=(1, 2))

#     areas_y = np.count_nonzero(patches_y, axis=(1, 2))
#     areas_x = np.count_nonzero(patches_x, axis=(1, 2))

#     ious = unions / (areas_x + areas_y - unions + 1e-8)

#     mask_ious = box_ious
#     mask_ious[(cy, cx)] = ious

#     return mask_ious


# def non_max_suppress_predictions(pred: DataDict, iou_threshold: float = 0.6) -> Array:
#     """Perform nms on the model prediction based on mask IOU

#     Args:
#         pred: A model prediction dictionary
#         ious_threshold: default 0.6

#     Returns:
#         mask: boolean mask of cells not supressed
#     """

#     ious = ious_of_patches_from_same_image(pred)
#     mask = ious > iou_threshold

#     cnt = 0
#     while np.count_nonzero(mask) != cnt:
#         cnt = jnp.count_nonzero(mask)
#         can_suppress_others = ~mask.any(axis=0)
#         suppressed = (mask & can_suppress_others[:, None]).any(axis=0)
#         mask = mask & ~suppressed[:, None]

#     return ~mask.any(axis=0)


_sampling_op = jax.vmap(partial(sub_pixel_samples, edge_indexing=True))


def rescale_patches(
    pred: DataDict,
    scale: float,
) -> tuple[Array, Array, Array]:
    """Rescale/resize instance outputs in a sub-pixel accurate way.
    If the input image was rescaled, this function take care of rescaling the predictions
    to the orginal coodinates.

    Args:
        pred: A model prediction dictionary
        scale: The scaling value. The function does not take a noop shortcut even if scale is 1.

    Returns: A tuple of three arrays
        patches: a 3D array of the rescaled segmentation patches. The array shape should be different from
            the orignal patches in model predition.
        yy: The y-coordinates of the patches in mesh-grid format
        xx: The x-coordinates of the patches in mesh-grid format
    """
    new_y0 = (pred["instance_yc"][:, 0, 0] + 0.5) * scale  # edge indexing in new scale
    new_x0 = (pred["instance_xc"][:, 0, 0] + 0.5) * scale  # edge indexing in new scale
    patch = pred["instance_output"]
    ps = round(patch.shape[-1] * scale)

    yy, xx = jnp.mgrid[:ps, :ps]
    yy = (
        jnp.floor(new_y0[:, None, None]).astype(int) + yy
    )  # center indexing in new scale
    xx = (
        jnp.floor(new_x0[:, None, None]).astype(int) + xx
    )  # center indexing in new scale

    # edge indexing in old scale
    rel_yy = (yy + 0.5) / scale
    rel_yy -= pred["instance_yc"][:, :1, :1]
    rel_xx = (xx + 0.5) / scale
    rel_xx -= pred["instance_xc"][:, :1, :1]

    new_patch = _sampling_op(
        patch,
        jnp.stack([rel_yy, rel_xx], axis=-1),
    )

    return new_patch, yy, xx



import dataclasses
import re
from typing import Any, Dict, Iterable, Optional, Set, Tuple

import jax
from flax import struct

from ..typing import *

InputLike = Any


def _unique_name(
    names: Set[str],
    name: str,
):

    if name in names:

        match = re.match(r"(.*?)(\d*)$", name)
        assert match is not None

        name = match[1]
        num_part = match[2]

        i = int(num_part) if num_part else 2
        str_template = f"{{name}}{{i:0{len(num_part)}}}"

        while str_template.format(name=name, i=i) in names:
            i += 1

        name = str_template.format(name=name, i=i)

    names.add(name)
    return name


def _unique_names(
    names: Iterable[str],
    *,
    existing_names: Optional[Set[str]] = None,
) -> Iterable[str]:
    if existing_names is None:
        existing_names = set()

    for name in names:
        yield _unique_name(existing_names, name)


def _lower_snake_case(s: str) -> str:
    s = re.sub(r"(?<!^)(?=[A-Z])", "_", s).lower()
    parts = s.split("_")
    output_parts = []

    for i in range(len(parts)):
        if i == 0 or len(parts[i - 1]) > 1:
            output_parts.append(parts[i])
        else:
            output_parts[-1] += parts[i]

    return "_".join(output_parts)


def _get_name(obj) -> str:
    if hasattr(obj, "name") and obj.name:
        return obj.name
    elif hasattr(obj, "func") and obj.func.__name__:
        return _lower_snake_case(obj.func.__name__)
    elif hasattr(obj, "__name__") and obj.__name__:
        return _lower_snake_case(obj.__name__)
    elif hasattr(obj, "__class__") and obj.__class__.__name__:
        return _lower_snake_case(obj.__class__.__name__)
    else:
        raise ValueError(f"Could not get name for: {obj}")


def list_to_tuple(maybe_list):
    if isinstance(maybe_list, list):
        return tuple(maybe_list)
    return maybe_list


class Inputs(struct.PyTreeNode):
    args: Tuple[Any, ...] = ()
    kwargs: Dict[str, Any] = dataclasses.field(default_factory=dict)

    def update(self, *args, **kwargs):
        tmp = self.kwargs.copy()
        tmp.update(kwargs)
        new_inputs = self.replace(args=self.args + args, kwargs=tmp)
        return new_inputs

    @classmethod
    def from_value(cls, value: InputLike) -> "Inputs":
        if isinstance(value, cls):
            return value
        elif isinstance(value, tuple):
            return cls(args=value)
        elif isinstance(value, dict):
            return cls(kwargs=value)
        else:
            return cls(args=(value,))


try:
    from torch.utils.data import DataLoader

    class TorchDataLoaderAdapter:
        """Convert torch dataloader into a python iterable suitable for [lacss.Trainer](./#lacss.train.base_trainer.Trainer)

        ```
        my_dataset = TorchDataLoaderAdapter(my_torch_dataloader)
        ```

        """

        def __init__(
            self,
            x: DataLoader,
            steps: int = -1,
        ):
            self.steps = steps
            self._dataset = x
            self.current_step = 0

        def get_dataset(self):
            def parse_dataloader_gen():
                self.current_step = 0
                for batch in iter(self._dataset):
                    self.current_step += 1
                    batch = jax.tree_util.tree_map(
                        lambda x: x.cpu().numpy(), list_to_tuple(batch)
                    )
                    yield batch

            return parse_dataloader_gen

        def __iter__(self):
            g = self.get_dataset()
            return g()

except ImportError:
    TorchDataLoaderAdapter = None


try:
    from tensorflow.data import Dataset

    class TFDatasetAdapter:
        """Convert `tf.data.Dataset` into a python iterable suitable for [lacss.train.Trainer](./#lacss.train.base_trainerxtrain.Trainer)

        ```
        my_dataset = TFDatasetAdapter(my_tf_dataset)
        ```

        """

        def __init__(self, ds: Dataset, steps=-1):
            self._dataset = ds

        def get_dataset(self):
            def parse_tf_data_gen():
                for batch in iter(self._dataset):
                    batch = jax.tree_util.tree_map(lambda x: x.numpy(), batch)
                    yield batch

            return parse_tf_data_gen

        def __iter__(self):
            g = self.get_dataset()
            return g()

except ImportError:
    TFDatasetAdapter = None



from __future__ import annotations

from functools import partial

import jax
from jax.scipy.signal import convolve

jnp = jax.numpy

from ..typing import *


def sorbel_edges(image: ArrayLike) -> Array:
    """Returns a tensor holding Sobel edge maps.

    Examples:

        >>> image = random.uniform(key, shape=[3, 28, 28])
        >>> sobel = sobel_edges(image)
        >>> sobel_y = sobel[0, :, :, :] # sobel in y-direction
        >>> sobel_x = sobel[1, :, :, :] # sobel in x-direction

    Args:
        image: [n, h, w]

    Returns:
        Tensor holding edge maps for each channel. [2, n, h, w]
    """

    kernels = jnp.array(
        [[[-1, -2, -1], [0, 0, 0], [1, 2, 1]], [[-1, 0, 1], [-2, 0, 2], [-1, 0, 1]]]
    )

    sorbel_filter = jax.vmap(
        jax.vmap(
            partial(convolve, mode="valid"),
            in_axes=[0, None],
        ),
        in_axes=[None, 0],
    )

    image = jnp.pad(image, [[0, 0], [1, 1], [1, 1]])
    output = sorbel_filter(image, kernels)

    return output


def _retrieve_value_at(img, loc, out_of_bound_value=0):

    iloc = jnp.floor(loc).astype(int)
    res = loc - iloc

    offsets = jnp.asarray(
        [[(i >> j) % 2 for j in range(len(loc))] for i in range(2 ** len(loc))]
    )
    ilocs = jnp.swapaxes(iloc + offsets, 0, 1)

    weight = jnp.prod(res * (offsets == 1) + (1 - res) * (offsets == 0), axis=1)

    max_indices = jnp.asarray(img.shape)[: len(loc), None]
    values = jnp.where(
        (ilocs >= 0).all(axis=0) & (ilocs < max_indices).all(axis=0),
        jnp.swapaxes(img[tuple(ilocs)], 0, -1),
        out_of_bound_value,
    )

    value = (values * weight).sum(axis=-1)

    return value


def sub_pixel_samples(
    img: ArrayLike,
    locs: ArrayLike,
    out_of_bound_value: float = 0,
    edge_indexing: bool = False,
) -> Array:
    """Retrieve image values as non-integer locations by interpolation

    Args:
        img: Array of shape [D1,D2,..,Dk, ...]
        locs: Array of shape [d1,d2,..,dn, k]
        out_of_bound_value: optional float constant, defualt 0.
        edge_indexing: if True, the index for the first value in img is 0.5, otherwise 0. Default is False

    Returns:
        values: [d1,d2,..,dn, ...], float
    """

    loc_shape = locs.shape
    img_shape = img.shape
    d_loc = loc_shape[-1]

    if edge_indexing:
        locs = locs - 0.5

    img = img.reshape(img_shape[:d_loc] + (-1,))
    locs = locs.reshape(-1, d_loc)
    op = partial(_retrieve_value_at, out_of_bound_value=out_of_bound_value)

    values = jax.vmap(op, in_axes=(None, 0))(img, locs)
    out_shape = loc_shape[:-1] + img_shape[d_loc:]

    values = values.reshape(out_shape)

    return values


def sub_pixel_crop_and_resize(
    img: ArrayLike, bbox: ArrayLike, output_shape: Shape, out_of_bound_value: float = 0
) -> Array:
    """Retrieve image values of a bbox. Resize output to output_shape. Used for ROI-Align.

    Args:
        img: Array of shape [H, W, ...]
        bbox: [y0, x0, y1, x1]
        output_shape: [h, w]
        out_of_bound_value: optional float constant, defualt 0.

    Returns:
        values: [h, w, ...], float
    """

    bbox = jnp.asarray(bbox)
    img = jnp.asarray(img)
    y0, x0, y1, x1 = bbox
    h, w = output_shape

    dy = (y1 - y0) / h
    dx = (x1 - x0) / w

    yy, xx = jnp.mgrid[(y0 + dy / 2) : y1 : (dy), (x0 + dx / 2) : x1 : dx]
    return sub_pixel_samples(
        img,
        jnp.stack([yy, xx], axis=-1),
        out_of_bound_value=out_of_bound_value,
        edge_indexing=True,
    )



"""
metrics classes for computing coco-style AP metrics.
"""

from __future__ import annotations

import jax
import jax.numpy as jnp
import numpy as np

from ..ops import *
from ..typing import *
from ..utils import unpack_x_y_sample_weight

"""
BE CAREFUL making changes here. Very easy to make a mistake and resulting mismatch with
coco algorithm. Be sure to validate against coco-evaluation before commit changes.
"""


def _unique_location_matching(similarity_matrix, threshold):
    """Perform matching based on similarity_matrix.

    This is different from the matchre functions in ops/*.py in that the matching
    will be unique: each column will match at most one row
    Args:
      similarity_matrix: [N, K] tensor, rows are presorted based on scores
      threshold: minimal value to be considered as a match
    Returns:
      matches: [N,] indices[0..K)  of the match for each row
      indicators: [N,] indicator (1/0) value for each match
    """

    matches = []
    indicators = []
    similarity_matrix = np.copy(similarity_matrix)
    n, _ = similarity_matrix.shape
    for i_row in range(n):
        row = similarity_matrix[i_row]
        row_max_indice = row.argmax()
        row_max = row[row_max_indice]
        matches.append(row_max_indice)
        if row_max > threshold:
            indicators.append(1)
            similarity_matrix[:, row_max_indice] = -1
        else:
            indicators.append(0)

    return np.array(matches, np.int32), np.array(indicators, np.int32)


def np_compute_ap(similarity_matrix, thresholds):
    # avoid edge cases
    _, k = similarity_matrix.shape
    if k == 0:
        return np.zeros([len(thresholds)], np.float32)

    apmks = []
    for th in thresholds:
        _, indicators = _unique_location_matching(similarity_matrix, th)
        p_k = np.cumsum(indicators) / (np.arange(len(indicators)) + 1)
        apmks.append(np.sum(p_k * indicators) / k)

    return np.array(apmks, np.float32)


class AP:
    """compute AP based on similarity_matrix and score.
    Usage:
      m = MeanAP([threshold_1, threshold_2,...])
      m.update(similarity_matrix, scores)
      ...
      m.compute()
    """

    def __init__(self, thresholds=[0.5], coco_style=False, name=None):
        self.thresholds = thresholds
        self.coco_style = coco_style
        if name is not None:
            self.name = name
        self.reset()

    def update(self, sm, scores):
        self.cell_counts += sm.shape[1]
        self.scores.append(scores)
        for th, indicators in zip(self.thresholds, self.indicator_list):
            _, ind = _unique_location_matching(sm, th)
            indicators.append(ind)

        self._result = None

    def compute(self):
        if self._result is not None:
            return self._result

        scores = np.concatenate(self.scores)
        indices = np.argsort(scores)
        aps = []
        for indicators in self.indicator_list:
            indicators = np.concatenate(indicators)
            indicators = indicators[indices[::-1]]
            p_k = np.cumsum(indicators) / (np.arange(len(indicators)) + 1)
            if self.coco_style:
                p_k = np.maximum.accumulate(p_k[::-1])[::-1]
            aps.append(np.sum(p_k[indicators == 1]) / self.cell_counts)

        self._result = np.array(aps, dtype=float)

        return self._result

    def reset(self):
        self.scores = []
        self.cell_counts = 0
        self.indicator_list = [[] for _ in range(len(self.thresholds))]
        self._result = np.array([-1.0] * len(self.thresholds))


class LoiAP(AP):
    """Compute AP of LACSS model based on location matching

    Usage:
      m = LoiAP([distance_threshold_1, distance_threshold_2,...])
      m.update(model_predication, gt_point_labels)
      ...
      m.compute()
    """

    def __init__(self, thresholds=[0.5], **kwargs):
        super().__init__([1 / th / th for th in thresholds], **kwargs)

    def update(self, batch, prediction):
        preds = prediction
        inputs, labels, _ = unpack_x_y_sample_weight(batch)
        if "gt_locations" in inputs:
            gt_locations = inputs["gt_locations"]
        else:
            gt_locations = labels["gt_locations"]

        score = np.asarray(preds["pred_scores"])
        pred = np.asarray(preds["pred_locations"])
        gt = np.asarray(gt_locations)

        row_mask = score > 0
        col_mask = (gt >= 0).all(axis=-1)
        dist2 = ((pred[:, None, :] - gt[:, :]) ** 2).sum(axis=-1)

        dist2 = 1.0 / (dist2 + 1e-8)
        dist2 = dist2[row_mask][:, col_mask]
        score = np.asarray(score)[row_mask]

        super().update(dist2, score)


class BoxAP(AP):
    """Compute bounding-box AP

    Usage:
      m = BoxAP([iou_threshold_1, iou_threshold_2,...])
      m.update(model_predication, gt_bboxes)
      ...
      m.compute()
    """

    def update(self, batch, prediction):
        preds = prediction
        _, labels, _ = unpack_x_y_sample_weight(batch)
        gt_bboxes = labels["gt_bboxes"]

        pred_bboxes = np.asarray(bboxes_of_patches(preds))
        gt_bboxes = np.asarray(gt_bboxes)
        iou = box_iou_similarity(pred_bboxes, gt_bboxes)
        score = np.asarray(preds["pred_scores"])
        gt_is_valid = (gt_bboxes >= 0).all(axis=-1)

        valid_preds = score >= 0
        iou = iou[valid_preds][:, gt_is_valid]
        score = score[valid_preds]
        super().update(iou, score)



from __future__ import annotations

from typing import Optional, Sequence

import flax.linen as nn
import jax
import jax.numpy as jnp

from ..typing import ArrayLike, DataDict
from .common import *
from .unet import UNet


class LacssCollaborator(nn.Module):
    """Collaborator module for semi-supervised Lacss training

    Attributes:
        conv_spec: conv-net specificaiton for cell border predicition
        unet_spec: specification for unet, used to predict cell foreground
        patch_size: patch size for the unet
        n_cls: number of classes (cell types) of input images
    """

    conv_spec: Sequence[int] = (32, 32)
    unet_spec: Sequence[int] = (16, 32, 64)
    patch_size: int = 1
    n_cls: int = 1

    @nn.compact
    def __call__(
        self, image: ArrayLike, cls_id: Optional[ArrayLike] = None
    ) -> DataDict:
        assert cls_id is not None or self.n_cls == 1
        c = cls_id.astype(int).squeeze() if cls_id is not None else 0

        net = UNet(self.unet_spec, self.patch_size)
        _, unet_out = net(image)

        y = unet_out[str(net.start_level)]

        fg = nn.Conv(self.n_cls, (3, 3))(y)
        fg = fg[..., c]

        if fg.shape != image.shape[:-1]:
            fg = jax.image.resize(fg, image.shape[:-1], "linear")

        y = image
        for n_features in self.conv_spec:
            y = nn.Conv(n_features, (3, 3), use_bias=False)(y)
            y = nn.GroupNorm(num_groups=None, group_size=1, use_scale=False)(
                y[None, ...]
            )[0]
            y = jax.nn.relu(y)

        y = nn.Conv(self.n_cls, (3, 3))(y)
        cb = y[..., c]

        return dict(
            fg_pred=fg,
            edge_pred=cb,
        )



from __future__ import annotations

from typing import Sequence, Union

import jax.numpy as jnp
from flax import struct

from ..typing import LossFunc
from .utils import _get_name


class LossLog(struct.PyTreeNode):
    loss_fn: LossFunc = struct.field(pytree_node=False)
    weight: jnp.ndarray = 1.0
    cnt: jnp.ndarray = 0.0
    sum: jnp.ndarray = 0.0

    def update(self, **kwargs):
        loss = self.loss_fn(**kwargs)
        if loss is None:
            return 0.0, self
        else:
            loss *= self.weight
            return loss, self.replace(cnt=self.cnt + 1, sum=self.sum + loss)

    def compute(self):
        return self.sum / self.cnt



from __future__ import annotations

from functools import partial

import jax
import jax.numpy as jnp
import optax

from ..ops import sub_pixel_samples
from ..utils import unpack_x_y_sample_weight
from .common import binary_focal_factor_loss, mean_over_boolean_mask


def supervised_instance_loss(batch, prediction):
    """LACSS instance loss, supervised with segmentation label"""
    preds = prediction
    _, labels, _ = unpack_x_y_sample_weight(batch)

    instance_mask = preds["instance_mask"]
    instance_logit = preds["instance_logit"]
    yc = preds["instance_yc"]
    xc = preds["instance_xc"]

    if not isinstance(labels, dict):
        labels = dict(gt_labels=labels)

    if "gt_labels" in labels:  # labeled with image label
        gt_labels = labels["gt_labels"].astype("int32")

        n_patches, ps, _ = yc.shape

        gt_labels = jnp.pad(gt_labels, ps // 2)
        gt_patches = gt_labels[yc + ps // 2, xc + ps // 2] == (
            jnp.arange(n_patches)[:, None, None] + 1
        )
        gt_patches = gt_patches.astype("float32")

    else:  # labeled with bboxes and rescaled segmentation masks, ie, coco
        y0, x0, y1, x1 = jnp.swapaxes(labels["gt_bboxes"], 0, 1)
        gt_segs = labels["gt_masks"]
        if len(gt_segs.shape) == 4:  # either NxHxWx1 or NxHxW
            gt_segs = gt_segs.squeeze(-1)
        seg_size = gt_segs.shape[1]

        # pixel size of the gt mask labels
        hs = (y1 - y0) / seg_size
        ws = (x1 - x0) / seg_size

        # compute rescaled coorinats in edge indexing
        yc = (yc - y0[:, None, None] + 0.5) / hs[:, None, None]
        xc = (xc - x0[:, None, None] + 0.5) / ws[:, None, None]

        # resample the label to match model coordinates
        gt_patches = jax.vmap(sub_pixel_samples)(
            gt_segs,
            jnp.stack([yc, xc], axis=-1) - 0.5,  # default is center indexing
        )
        gt_patches = (gt_patches >= 0.5).astype("float32")

    loss = optax.sigmoid_binary_cross_entropy(instance_logit, gt_patches)

    return mean_over_boolean_mask(loss, instance_mask)


def self_supervised_instance_loss(batch, prediction, *, soft_label: bool = True):
    """Unsupervised instance loss"""
    preds = prediction
    _, labels, _ = unpack_x_y_sample_weight(batch)

    instance_mask = preds["instance_mask"]
    instances = preds["instance_output"]
    instance_logit = preds["instance_logit"]
    yc = preds["instance_yc"]
    xc = preds["instance_xc"]

    patch_size = instances.shape[-1]
    padding_size = patch_size // 2 + 2
    yc += padding_size
    xc += padding_size

    binary_mask = jax.lax.stop_gradient(jax.nn.sigmoid(preds["fg_pred"]))
    seg = jnp.pad(binary_mask, padding_size)

    if soft_label:
        seg_patch = seg[yc, xc]

        loss = (1 - seg_patch) * instances + seg_patch * (1 - instances)

        instance_sum = jnp.zeros_like(seg)
        instance_sum = instance_sum.at[yc, xc].add(instances)
        instance_sum_i = instance_sum[yc, xc] - instances

        loss += instances * instance_sum_i

    else:
        seg = (seg > 0.5).astype(instances.dtype)
        seg_patch = seg[yc, xc]

        loss = (1 - seg_patch) * instances + seg_patch * (1 - instances)

        log_yi_sum = jnp.zeros_like(seg)
        log_yi = -jax.nn.log_sigmoid(-instance_logit)
        log_yi_sum = log_yi_sum.at[yc, xc].add(log_yi)
        log_yi = log_yi_sum[yc, xc] - log_yi

        loss = loss + (instances * log_yi)

    return mean_over_boolean_mask(loss, instance_mask)


def weakly_supervised_instance_loss(batch, prediction, *, ignore_mask: bool = False):
    """Instance loss supervised by image mask instead of instance masks"""
    preds = prediction
    inputs, labels, _ = unpack_x_y_sample_weight(batch)

    instance_mask = preds["instance_mask"]
    instances = preds["instance_output"]
    instance_logit = preds["instance_logit"]
    yc = preds["instance_yc"]
    xc = preds["instance_xc"]

    patch_size = instances.shape[-1]
    padding_size = patch_size // 2 + 2
    yc += padding_size
    xc += padding_size

    if ignore_mask:
        seg = jnp.zeros(inputs["image"].shape[:-1])
        seg = jnp.pad(seg, padding_size)
        loss = jnp.zeros_like(instances)
    else:
        if isinstance(labels, dict):
            seg = labels["gt_image_mask"].astype("float32")
        else:
            seg = labels.astype("float32")
        seg = jnp.pad(seg, padding_size)
        seg_patch = seg[yc, xc]
        loss = (1.0 - seg_patch) * instances + seg_patch * (1.0 - instances)

    log_yi_sum = jnp.zeros_like(seg)

    log_yi = -jax.nn.log_sigmoid(-instance_logit)
    log_yi_sum = log_yi_sum.at[yc, xc].add(log_yi)
    log_yi = log_yi_sum[yc, xc] - log_yi

    loss = loss + (instances * log_yi)

    return mean_over_boolean_mask(loss, instance_mask)


def segmentation_loss(batch, prediction, *, pretraining=False):
    preds = prediction
    inputs, labels, _ = unpack_x_y_sample_weight(batch)

    if labels is None:
        labels = {}

    if "gt_labels" in labels or "gt_masks" in labels:  # supervised
        return supervised_instance_loss(batch, prediction)
    elif "gt_image_mask" in labels:  # supervised by point + imagemask
        return weakly_supervised_instance_loss(batch, prediction)
    else:  # point-supervised
        return self_supervised_instance_loss(
            batch, prediction, soft_label=not pretraining
        )



import dataclasses

import jax.numpy as jnp
import numpy as np
from flax.core.frozen_dict import freeze, unfreeze


def _to_str(p):
    return "".join(p.astype(int).reshape(-1).astype(str).tolist())


def format_predictions(pred, mask=None, encode_patch=True, threshold=0.5):
    """Produce more readable data from model predictions
    Args:
        pred: model output without batch dim
        mask: optional mask selecting cells
        threshold: float patch threshold
    Returns: dict(locations, scores, centroids, bboxes, encodings)
    """
    from lacss.ops import bboxes_of_patches

    patches = pred["instance_output"] >= threshold
    yc = pred["instance_yc"]
    xc = pred["instance_xc"]

    bboxes = bboxes_of_patches(pred)

    n_pixels = np.count_nonzero(patches, axis=(1, 2))
    centroids = jnp.stack(
        [
            (patches * yc).sum(axis=(1, 2)) / n_pixels,
            (patches * xc).sum(axis=(1, 2)) / n_pixels,
        ],
        axis=-1,
    )  # centroid

    outputs = dict(
        locations=pred["pred_locations"],
        scores=pred["pred_scores"],
        centroids=centroids,
        bboxes=bboxes,
    )

    is_valid = pred["instance_mask"].squeeze(axis=(-1, -2))
    is_valid &= patches.any(axis=(1, 2))  # no empty patches
    if mask is not None:
        is_valid &= mask

    outputs = {k: np.asarray(v)[is_valid] for k, v in outputs.items()}

    if encode_patch:

        encodings = []
        patches = np.asarray(patches)[is_valid]
        y0 = np.asarray(yc)[is_valid, 0, 0]
        x0 = np.asarray(xc)[is_valid, 0, 0]
        boxes = np.asarray(outputs["bboxes"]) - np.stack([y0, x0, y0, x0], axis=-1)

        for box, patch in zip(boxes, patches):
            roi = patch[box[0] : box[2], box[1] : box[3]]
            encodings.append(_to_str(roi))

        outputs["encodings"] = encodings

    return outputs


def show_images(imgs, locs=None, **kwargs):
    import matplotlib.patches
    import matplotlib.pyplot as plt

    fig, axs = plt.subplots(1, len(imgs), figsize=(4 * len(imgs), 5))
    if len(imgs) == 1:
        axs = [axs]

    for k, img in enumerate(imgs):
        axs[k].imshow(img, **kwargs)
        axs[k].axis("off")
        if locs is not None and locs[k] is not None:
            loc = np.round(locs[k]).astype(int)
            for p in loc:
                c = matplotlib.patches.Circle(
                    (p[1], p[0]), fill=False, edgecolor="white"
                )
                axs[k].add_patch(c)
    plt.tight_layout()


def dataclass_from_dict(klass, dikt):
    try:
        fieldtypes = {f.name: f.type for f in dataclasses.fields(klass)}
        return klass(**{f: dataclass_from_dict(fieldtypes[f], dikt[f]) for f in dikt})
    except:
        return dikt


def load_from_pretrained(pretrained: str):
    """Load a saved model.

    Args:
        pretrained: The url to the saved model.

    Returns: A tuple (module, parameters) representing the model.
    """
    import os

    if os.path.isdir(pretrained):
        # directory are orbax checkpoint

        from lacss.train import LacssTrainer

        trainer = LacssTrainer.from_checkpoint(pretrained)
        module = trainer.model.principal
        params = trainer.params["principal"]

    else:
        # uri or files were treated as pickled byte steam
        import pickle

        from .modules import Lacss
        from .train import Trainer

        if os.path.isfile(pretrained):
            with open(pretrained, "rb") as f:
                thingy = pickle.load(f)

        else:
            from urllib.request import Request, urlopen

            headers = {"User-Agent": "Wget/1.13.4 (linux-gnu)"}
            req = Request(url=pretrained, headers=headers)

            bytes = urlopen(req).read()
            thingy = pickle.loads(bytes)

        if isinstance(thingy, Trainer):
            module = thingy.model
            params = thingy.params

        else:
            cfg, params = thingy

            if isinstance(cfg, Lacss):
                module = cfg
            else:
                module = Lacss.from_config(cfg)

    if "params" in params and len(params) == 1:
        params = params["params"]

    return module, freeze(params)


def pack_x_y_sample_weight(x, y=None, sample_weight=None):
    """Packs user-provided data into a tuple."""
    if y is None:
        return (x,)
    elif sample_weight is None:
        return (x, y)
    else:
        return (x, y, sample_weight)


def unpack_x_y_sample_weight(data):
    """Unpacks user-provided data tuple."""
    if not isinstance(data, tuple):
        return (data, None, None)
    elif len(data) == 1:
        return (data[0], None, None)
    elif len(data) == 2:
        return (data[0], data[1], None)
    elif len(data) == 3:
        return (data[0], data[1], data[2])

    raise ValueError("Data not understood.")

def make_label_continuous(label, dtype=None):
    """ Relabel a label image so that the label values are continuous. It is assumed that the
    label starts with 0. If there is negative numbers in the input, they will be replaced by 0.

    Args:
        label: input label image
    
    Keyword Args:
        dtype: the dtype of the output image. default is to keep the dtype of the input.
    
    Returns:
        Relabeled image.
    """
    if not isinstance(label, np.ndarray) and dtype is None:
        raise ValueError("A dtype must be specified if the input data is not a np array")
    elif dtype is None:
        dtype = label.dtype

    label = np.asarray(label, dtype=dtype)

    k = np.unique(label)
    v = np.asarray(range(len(k)))

    mapping_ar = np.zeros(k.max() + 1, dtype=dtype)
    mapping_ar[k] = v

    return mapping_ar[label]




from __future__ import annotations

import pathlib
import pickle
import warnings
from functools import lru_cache, partial
from pathlib import Path
from typing import Any, Iterable, Iterator, Optional, Sequence, Union

import flax.linen as nn
import jax
import jax.numpy as jnp
import numpy as np
import optax
from flax.core.frozen_dict import freeze, unfreeze
from flax.training.train_state import TrainState

from lacss.utils import unpack_x_y_sample_weight

from ..typing import *
from .loss import LossLog
from .strategy import JIT
from .utils import Inputs, _get_name

WeightedLossFunc = LossFunc | tuple[LossFunc, float | np.number]
LOSSES = Union[LossFunc, Sequence[LossFunc]]
METRICS = Union[Metric, Sequence[Metric]]

# so that multiple calls return the same obj
# this avoids JIT when supplying partial func as args
_cached_partial = lru_cache(partial)


class Trainer:
    """A general purpose FLAX model trainer. Help avoiding most of the biolerplate code when trainning with FLAX.

    Attributes:
        model: A Flax module
        losses: A list of loss function (or other callabels).
        optimizer: An optax optimizer
        seed: RNG seed
        params: Current model parameters. This is a frozen dict. This is read/write.
        initialized: Whether the model has been initialized with an optimizer and initial weights.

    Example:

        ```
        trainer = lacss.train.Trainer(my_module, my_loss_func)

        trainer.initialize(my_dataset, tx=my_optimzier)

        train_it = trainer.train(my_dataset)

        for k in range(train_steps):
            loss_logs = next(train_it)
            if k % 1000 == 0:
                print(loss_logs)
                trainer.reset()
        ```

    """

    def __init__(
        self,
        model: nn.Module,
        losses: LOSSES | None = None,
        optimizer: Optional[Optimizer] = None,
        seed: Union[int, Array] = 42,
        strategy: type = JIT,
        *,
        parameters: dict | None = None,
    ):
        """Constructor

        Args:
            model: A Flax module. It can be a bound module with parameters.
            losses: A loss function (callabels) or a list of loss functions, or None.
                These should have the call signiture of:
                    loss = loss_fn(inputs, labels, preds)
                The "preds" is the model output. The "inputs" and "labels" are from the train dataset.
                The model trains on the sum of all losses.
            optimizer: An optax optimzier
            seed: RNG seed.
            strategy: Training backend. See [Traing backends](./#training-backends).

        """
        if model.scope is not None:
            model, variables = model.unbind()
            if parameters is None and "params" in variables:
                parameters = variables["params"]

        self.model = model
        self.losses = losses

        self.seed = seed if isinstance(seed, jnp.ndarray) else jax.random.PRNGKey(seed)
        self.optimizer = optimizer

        self._strategy = strategy

        self.parameters = parameters

    def reset(
        self,
        loss_weights: Optional[Sequence[float]] = None,
        losses: Optional[LOSSES] = None,
    ):
        """Reset internal loss value tracking

        Args:
            loss_weights: Optional weights of individual loss functions. If not None, the
                total loss is weighted sum.
            losses: Optionally override the class losses defined as the object initialization.
        """
        if losses is None:
            losses = self.losses

            if losses is None:
                raise ValueError(f"No loss functions provided")

        try:
            iter(losses)
        except:
            losses = (losses,)

        if loss_weights is None:
            loss_weights = (1.0,) * len(losses)

        if len(loss_weights) != len(losses):
            raise ValueError(
                f"Loss weights supplied {loss_weights} does not match the number of loss functions ({len(losses)})"
            )

        self.loss_logs = tuple(
            LossLog(loss, w) for loss, w in zip(losses, loss_weights)
        )

    @property
    def initialized(self):
        return self.parameters is not None

    def initialize(self, data: Iterable, tx: Optional[Optimizer] = None) -> None:
        """Generate initial model weights. Does nothing if self.parameters is not None.

        Args:
            data: An iterable to produce training dataset.
                see [train()](./#lacss.train.base_trainer.Trainer.train)
            tx: Depreciated. Optional optax optimzier for when the object was constructed without an
                optimizer
        """
        if tx is not None:
            self.optimizer = tx
            warnings.warn(
                f"Using tx parameter in initialize() call is deprecated",
                DeprecationWarning,
            )

        if self.parameters is None:
            peek = next(iter(data))
            inputs, _, _ = unpack_x_y_sample_weight(peek)
            # inputs = peek[0] if isinstance(peek, tuple) else peek

            self.seed, key = jax.random.split(self.seed)
            variables = self._strategy.init_fn(key, self.model, inputs)
            self.parameters = variables["params"]

        # self.state = TrainState.create(
        #     apply_fn=self.model.apply,
        #     params=variables["params"],
        #     tx=tx,
        # )

    def __call__(
        self, inputs: Any, *, strategy: Optional[type] = None, **kwargs
    ) -> Any:
        """Calling the underlying model

        Args:
            inputs: A tuple or a dict as the inputs for the model.
            strategy: Optionally override the default strategy.
            **kwargs: Additional keyword args passed on to the model

        Returns:
            Model outputs.
        """
        if strategy is None:
            strategy = self._strategy

        predict_fn = strategy.predict

        apply_fn = _cached_partial(self.model.apply, **kwargs)
        preds = predict_fn(apply_fn, self.parameters, inputs)

        return preds

    def compute_loss_log(self) -> dict:
        return {
            _get_name(loss_log.loss_fn): loss_log.compute()
            for loss_log in self.loss_logs
        }

    def train(
        self,
        dataset: Iterable,
        strategy: Optional[type] = None,
        rng_cols: Sequence[str] = None,
        frozen: dict | None = None,
        **kwargs,
    ) -> Iterator:
        """Create the training iterator

        Args:
            dataset: An iterator or iterable to supply the training data.
                The dataset should produce ```(inputs, labels, sample_weight)```, however
                both the labels and the sample_weight are optional. The inputs is either a tuple
                or a dict. If the inputs is a dict, the keys are interpreted as the names for
                keyword args of the model's __call__ function.
            strategy: Optionally override the default strategy.
            rng_cols: Names of any RNG used by the model. Should be a list of strings.
            **kwargs: Additional keyward args passed to the model. E.g. "training=True"

        Returns:
            A iterator. Stepping through the iterator will train the model. The iterator
                itself returns a loss log dict, which are mean loss values for each loss
                function.

        """
        if strategy is None:
            strategy = self._strategy

        if self.parameters is None:
            self.initialize(dataset)

        assert self.parameters is not None

        if frozen is None:
            frozen = jax.tree_util.tree_map(lambda _: False, self.parameters)

        optimizers = {True: optax.set_to_zero(), False: self.optimizer}
        tx = optax.multi_transform(
            optimizers,
            frozen,
        )

        train_state = TrainState.create(
            apply_fn=_cached_partial(self.model.apply, **kwargs),
            params=self.parameters,
            tx=tx,
        )

        train_fn = strategy.train_step

        self.reset()

        self.seed, seed = jax.random.split(self.seed)

        for batch in dataset:
            if rng_cols is not None:
                key = jax.random.fold_in(seed, train_state.step)
                keys = jax.random.split(key, len(rng_cols))
                rngs = {name: k for name, k in zip(rng_cols, keys)}
            else:
                rngs = None

            train_state, self.loss_logs, preds = train_fn(
                train_state, self.loss_logs, batch, rngs
            )

            batch_logs = self.compute_loss_log()

            self.parameters = self.parameters.copy(train_state.params)

            yield batch_logs

    def save_model(self, path: PathLike, sub_module: Optional[str] = None) -> None:
        """Save the model in a pickled file. The pickle is a tuple of
            (module, weights).

        Args:
            path: The file path.
            sub_module: Optionally only save a sub_module of the model
                by specifying the name
        """
        module = self.model
        params = self.parameters

        if sub_module is not None:
            module = module.bind(dict(params=params))
            module = getattr(module, sub_module)
            module, params = module.unbind()
            params = params["params"]

        if isinstance(path, str):
            path = Path(path)

        path.parent.mkdir(parents=True, exist_ok=True)

        with open(path, "wb") as f:
            pickle.dump((module, params), f)

    # def pickle_self(self, path: PathLike) -> None:
    #     """Make a pickle save of the trainer. This saves the model as well as
    #     the training states.

    #     Args:
    #         path: The file path.
    #     """
    #     path = pathlib.Path(path)
    #     path.parent.mkdir(parents=True, exist_ok=True)

    #     with open(path, "wb") as f:
    #         pickle.dump(self, f)

    # @staticmethod
    # def restore_from_pickle(path: PathLike):
    #     """Restore from the a pickled saved.

    #     Args:
    #         path: The pickle file path.

    #     Returns:
    #         A new trainer object.
    #     """
    #     path = pathlib.Path(path)

    #     try:
    #         _bytes = path.read_bytes()
    #     except BaseException as e:
    #         raise OSError(f"Could not load the checkpoint. Got exception: {e}")

    #     trainer = pickle.loads(_bytes)

    #     if not isinstance(trainer, Trainer):
    #         raise TypeError("The saved obj is not a Trainer checkpoint")

    #     return trainer

    def test(
        self,
        dataset: Iterable,
        metrics: METRICS,
        strategy: Optional[type] = None,
        **kwargs,
    ) -> Iterator:
        """Create test/validation iterator.

        Args:
            dataset: An iterator or iterable to supply the testing data.
                The iterator should yield a tupple of (inputs, labels).
            metrics: A list of Metric objects. They should have two functions:
                m.update(preds, **kwargs):
                    preds is the model output. the remaining kwargs are content of
                    labels.
                m.compute():
                    which should return the accumulated metric value.
            strategy: Optionally override the default strategy.

        Returns:
            An iterator. Stepping through it will drive the updating of each metric
                obj. The iterator itself return the list of metrics.
        """
        if strategy is None:
            strategy = self._strategy

        try:
            iter(metrics)
        except TypeError:
            metrics = [metrics]

        apply_fn = _cached_partial(self.model.apply, **kwargs)

        predict_fn = strategy.predict

        for data in dataset:
            # inputs, labels, _ = unpack_x_y_sample_weight(data)
            inputs = data[0] if isinstance(data, tuple) else data
            preds = predict_fn(apply_fn, self.parameters, inputs)
            kwargs = dict(
                batch=data,
                prediction=preds,
            )
            for m in metrics:
                m.update(**kwargs)

            yield metrics

    def test_and_compute(self, *args, **kwargs) -> dict:
        """A convient function to compute all metrics. See [test() fucntion](./#lacss.train.base_trainer.Trainer.test)

        Returns:
            A metric dict. Keys are metric names.
        """
        for metrics in self.test(*args, **kwargs):
            pass
        return {_get_name(m): m.compute() for m in metrics}

    @property
    def parameters(self) -> Params:
        return self._params

    @parameters.setter
    def parameters(self, new_params: Params | None) -> None:
        if new_params is not None:
            self._params = freeze(new_params)
        else:
            self._params = None

    # @property
    # def optimizer(self) -> Optimizer:
    #     return self._optimizer

    # @optimizer.setter
    # def optimizer(self, tx: Optimizer) -> None:
    #     self._optimizer = tx

    # if self._initialized:
    #     self.state = TrainState.create(
    #         apply_fn=self.model.apply,
    #         params=self.params,
    #         tx=tx,
    #     )



from __future__ import annotations

import typing as tp
from functools import partial

import jax
import jax.numpy as jnp
import jax.tree_util as jtu
from flax.training.train_state import TrainState

from .loss import LossLog
from .utils import Inputs


def _tree_merge(tree_a, tree_b):
    param_dict = dict(jax.tree_util.tree_leaves_with_path(tree_a))
    for k, v in jax.tree_util.tree_leaves_with_path(tree_b):
        param_dict[k] = v
    return jax.tree_util.tree_unflatten(
        jax.tree_util.tree_structure(tree_a), param_dict.values()
    )


class Eager:
    @classmethod
    def loss_fn(cls, params, loss_logs, batch, rngs, apply_fn):
        inputs = batch[0] if isinstance(batch, tuple) else batch

        # if sample_weight is None:
        #     sample_weight = 1.0

        inputs_obj = Inputs.from_value(inputs)

        preds = apply_fn(
            {"params": params},
            *inputs_obj.args,
            **inputs_obj.kwargs,
            rngs=rngs,
        )

        args = dict(
            batch=batch,
            prediction=preds,
        )

        losses, loss_logs = zip(*[loss_log.update(**args) for loss_log in loss_logs])
        total_loss = sum(losses)

        return total_loss, (loss_logs, preds)

    @classmethod
    def init_fn(cls, key, model, inputs):
        inputs_obj = Inputs.from_value(inputs)

        state = model.init(key, *inputs_obj.args, **inputs_obj.kwargs)

        return state

    @classmethod
    def predict(cls, apply_fn, params, inputs):  # only if model is immutable
        # print("JIT Predict")
        inputs_obj = Inputs.from_value(inputs)
        preds = apply_fn({"params": params}, *inputs_obj.args, **inputs_obj.kwargs)
        return preds

    @classmethod
    def train_step(
        cls,
        state: TrainState,
        loss_logs: tp.Sequence[LossLog],
        batch: tp.Any,
        rngs: tp.Optional[dict],
    ) -> tp.Tuple[TrainState, tp.Sequence[LossLog], tp.Any]:
        # print('JIT train_step')
        grad_fn = jax.grad(
            partial(cls.loss_fn, apply_fn=state.apply_fn),
            has_aux=True,
        )
        grads, (loss_logs, preds) = grad_fn(
            state.params,
            loss_logs,
            batch,
            rngs,
        )
        state = state.apply_gradients(grads=grads)

        return state, loss_logs, preds


class Core(Eager):
    predict = jax.jit(Eager.predict, static_argnames="apply_fn")

    @classmethod
    def init_fn(cls, key, model, inputs):
        inputs_obj = Inputs.from_value(inputs)

        state = jax.jit(model.init)(key, *inputs_obj.args, **inputs_obj.kwargs)

        return state


class JIT(Core):
    train_step = jax.jit(Core.train_step)


class _Distributed(Eager):
    @classmethod
    def _train_step(
        cls,
        state: TrainState,
        loss_logs: tp.Sequence[LossLog],
        batch: tp.Any,
        rngs: tp.Optional[dict],
    ) -> tp.Tuple[TrainState, tp.Sequence[LossLog], tp.Any]:
        # print("JITTTTING")
        axis_index = jax.lax.axis_index("mapped")
        rngs = jax.tree_util.tree_map(
            lambda key: jax.random.fold_in(key, axis_index), rngs
        )
        grad_fn = jax.grad(
            partial(cls.loss_fn, apply_fn=state.apply_fn),
            has_aux=True,
        )

        grads, (loss_logs, preds) = grad_fn(
            state.params,
            loss_logs,
            batch,
            rngs,
        )

        grads = jax.lax.pmean(grads, axis_name="mapped")
        state = state.apply_gradients(grads=grads)

        # aggregate logs
        loss_logs = jax.tree_map(partial(jax.lax.pmean, axis_name="mapped"), loss_logs)

        #         # sync batch statistics
        #         model.map(partial(jax.lax.pmean, axis_name="mapped"), tx.BatchStat, inplace=True)

        return state, loss_logs, preds

    @classmethod
    def init_fn(cls, key, model, inputs):
        inputs = jax.tree_map(lambda v: v[0], inputs)
        return Eager.init_fn(key, model, inputs)


class Distributed(_Distributed):
    train_step = jax.pmap(
        _Distributed._train_step,
        axis_name="mapped",
        in_axes=(None, None, 0, None),
        out_axes=(None, None, 0),
    )

    predict = jax.pmap(
        Eager.predict,
        axis_name="mapped",
        in_axes=(None, None, 0),
        static_broadcasted_argnums=0,
    )


class VMapped(_Distributed):
    train_step = jax.jit(
        jax.vmap(
            _Distributed._train_step,
            axis_name="mapped",
            in_axes=(None, None, 0, None),
            out_axes=(None, None, 0),
        ),
    )

    predict = jax.jit(
        jax.vmap(
            Eager.predict,
            axis_name="mapped",
            in_axes=(None, None, 0),
        ),
        static_argnames="apply_fn",
    )



""" 
Attributes:
    model_urls: URLs for build-in pretrain models. e.g model_urls["default"].
"""
from __future__ import annotations

import logging
import pickle
from functools import lru_cache, partial, reduce
from typing import Mapping, Optional, Sequence, Tuple, Union

import cv2
import flax.linen as nn
import jax
import jax.numpy as jnp
import numpy as np
from flax.core.frozen_dict import freeze, unfreeze

import lacss.modules
import lacss.ops

from ..ops import patches_to_label, sorted_non_max_suppression
from ..typing import *
from ..utils import load_from_pretrained

Shape = Sequence[int]

_cached_partial = lru_cache(partial)

model_urls: Mapping[str, str] = {
    "cnsp4-bf": "https://huggingface.co/jiyuuchc/lacss-cnsp4-bf/resolve/main/cnsp4_bf.bin?download=true",
    "cnsp4-fl": "https://huggingface.co/jiyuuchc/lacss-cnsp4-fl/resolve/main/cnsp4_fl.bin?download=true",
    "cnsp4-base": "https://huggingface.co/jiyuuchc/lacss-cnsp4-base/resolve/main/cnsp4_base.bin?download=true",
}
model_urls["default"] = model_urls["cnsp4-base"]


def _to_polygons(
    pred,
    *,
    segmentation_threshold=0.5,
    scaling=1.0,
    chain_approx=cv2.CHAIN_APPROX_SIMPLE,
) -> list:
    polygons = []

    y0s = np.asarray(pred["y0s"])
    x0s = np.asarray(pred["x0s"])
    scores = np.asarray(pred["scores"])
    segs = np.asarray(pred["segmentations"] >= segmentation_threshold).astype("uint8")

    if "instance_mask" in pred:
        mask = np.asarray(pred["instance_mask"])
        segs = segs[mask]
        y0s = y0s[mask]
        x0s = x0s[mask]
        scores = scores[mask]

    for y0, x0, seg, score in zip(y0s, x0s, segs, scores):
        c, _ = cv2.findContours(seg, cv2.RETR_EXTERNAL, chain_approx)

        max_len_element = reduce(
            lambda a, b: a if len(a) >= len(b) else b,
            c,
            np.zeros([0, 1, 2], dtype=int),
        )

        polygon = max_len_element.squeeze(1).astype(float) + [x0, y0]

        polygons.append(polygon)

    if scaling != 1.0:
        polygons = [(c + 0.5) / scaling - 0.5 for c in polygons]

    return polygons


def _draw_label(polygons, image_size) -> np.ndarray:
    label = np.zeros(image_size, dtype="int32")
    color = len(polygons)
    for polygon in polygons[::-1]:
        if len(polygon) > 0:
            polygon = np.round(polygon).astype(int)
            cv2.fillPoly(label, [polygon], color)
        color -= 1

    return label


def _remove_edge_instances(
    pred: dict,
    img_shape: Tuple[int, int],
    remove_top: bool = True,
    remove_bottom: bool = True,
    remove_left: bool = True,
    remove_right: bool = True,
    threshold: float = 0.5,
    min_border_pixel: int = 1,
):
    h, w = img_shape
    segs = pred["segmentations"] >= threshold
    yc, xc = jnp.mgrid[: segs.shape[-2], : segs.shape[-1]]
    yc += pred["y0s"][:, None, None]
    xc += pred["x0s"][:, None, None]

    removal = jnp.zeros([len(segs)], dtype=bool)

    removal |= remove_top & (
        jnp.count_nonzero((yc <= 0) & segs, axis=(1, 2)) >= min_border_pixel
    )

    removal |= remove_bottom & (
        jnp.count_nonzero((yc >= h - 1) & segs, axis=(1, 2)) >= min_border_pixel
    )

    removal |= remove_left & (
        jnp.count_nonzero((xc <= 0) & segs, axis=(1, 2)) >= min_border_pixel
    )

    removal |= remove_right & (
        jnp.count_nonzero((xc >= w - 1) & segs, axis=(1, 2)) >= min_border_pixel
    )

    removal = removal.reshape(pred["instance_mask"].shape)

    pred["instance_mask"] &= ~removal

    return pred


@partial(jax.jit, static_argnums=0)
def _predict(context, params, image):
    scaling = context["scaling"]
    apply_fn = context["apply_fn"]
    output_type = context["output_type"]
    nms_iou = context["nms_iou"]
    h, w = image.shape[:2]

    if scaling != 1:
        image = jax.image.resize(
            image, [round(h * scaling), round(w * scaling), c], "linear"
        )

    preds = apply_fn(dict(params=params), image)

    instance_mask = preds["instance_mask"].squeeze(axis=(1, 2))
    instance_mask &= preds["pred_scores"] >= context["score_threshold"]

    if context["min_area"] > 0:
        areas = jnp.count_nonzero(
            preds["instance_output"] > context["segmentation_threshold"], axis=(1, 2)
        )
        instance_mask &= areas > context["min_area"]

    if context["remove_out_of_bound"]:
        pred_locations = preds["pred_locations"]
        h, w = image.shape[:2]
        valid_locs = (pred_locations >= 0).all(axis=-1)
        valid_locs &= pred_locations[:, 0] < h
        valid_locs &= pred_locations[:, 1] < w
        instance_mask &= valid_locs

    output = dict(
        instance_mask=instance_mask,
        segmentations=preds["instance_output"],
        y0s=preds["instance_yc"][:, 0, 0],
        x0s=preds["instance_xc"][:, 0, 0],
        scores=preds["pred_scores"],
    )

    need_to_comput_bbox = output_type == "bbox" or nms_iou > 0 or output_type == "grid"

    if need_to_comput_bbox:
        output["bboxes"] = lacss.ops.bboxes_of_patches(preds)

    if output_type == "bbox":
        output["segmentations"] = jax.vmap(_comput_masks)(
            output["segmentations"],
            output["y0s"],
            output["x0s"],
            output["bboxes"],
        )

    if nms_iou > 0 and output_type != "grid":  # need nms
        boxes = output["bboxes"]
        mask = output["instance_mask"]
        boxes = jnp.where(mask[:, None], boxes, -1)
        _, _, selections = sorted_non_max_suppression(
            output["scores"],
            boxes,
            -1,
            threshold=nms_iou,
            return_selection=True,
        )
        output["instance_mask"] &= selections

    # try to compute label
    if output_type == "label" and scaling < 2.0 and scaling > 0.5:
        if scaling != 1.0:
            (
                preds["instance_output"],
                preds["instance_yc"],
                preds["instance_xc"],
            ) = lacss.ops.rescale_patches(preds, 1 / scaling)

        label = lacss.ops.patches_to_label(
            preds,
            input_size=[h, w],
            mask=output["instance_mask"],
            score_threshold=0,
            threshold=context["segmentation_threshold"],
        )
        output.update(dict(label=label.astype(int)))

    return output


def _comput_masks(seg, y0, x0, bbox, mask_dim=[48, 48]):
    by0 = bbox[0] - y0
    bx0 = bbox[1] - x0
    bh = bbox[2] - bbox[0]
    bw = bbox[3] - bbox[1]

    mg = jnp.mgrid[: mask_dim[0], : mask_dim[1]].transpose((1, 2, 0)) + 0.5
    mg /= jnp.asarray(mask_dim)

    mg = mg * jnp.asarray([bh, bw]) + jnp.asarray([by0, bx0])

    return lacss.ops.sub_pixel_samples(seg, mg, edge_indexing=True)


class Predictor:
    """Main class interface for model deployment. This is the only class you
    need if you don't train your own model

    Examples:
        The most common use case is to use a build-in pretrained model.

            import lacss.deploy

            # look up the url of a build-in mode
            url = lacss.deploy.model_urls["default"]

            # create the predictor instance
            predictor = lacss.deploy.Predictor(url)

            # make a prediction
            label = predictor.predict(image)

    Attributes:
        module: The underlying FLAX module
        params: Model weights.
        detector: The detector submodule for convinence. A common use case
            is to customize this submodule during inference. e.g.
            ```
            predictor.detector.test_min_score = 0.5
            ```
            Detector submodule has no trained paramters.
    """

    def __init__(self, url: str, precompile_shape: Optional[Shape] = None):
        """Construct Predictor

        Args:
            url: A URL or local path to the saved model.
                URLs for build-in pretrained models can be found in lacss.deploy.model_urls
            precompile_shape: Image shape(s) for precompiling. Otherwise
                the model will be recompiled for every new input image shape.

        """

        if isinstance(url, tuple) and len(url) == 2:
            self.model = url
            if not isinstance(self.model[0], nn.Module):
                raise ValueError(
                    "Initiaize the Predictor with a tuple, but the first element is not a Module."
                )
        else:
            self.model = load_from_pretrained(url)

        if precompile_shape is not None:
            logging.info("Prcompile the predictor for image shape {precompile_shape}.")
            logging.info("This will take several minutes.")

            try:
                iter(precompile_shape[0])
            except:
                precompile_shape = [precompile_shape]
            for shape in precompile_shape:
                x = jnp.zeros(shape)
                _ = self.predict(x)

    def predict(
        self,
        image: ArrayLike,
        *,
        min_area: float = 0,
        remove_out_of_bound: bool = False,
        scaling: float = 1.0,
        output_type: str = "label",
        score_threshold: float = 0.5,
        segmentation_threshold: float = 0.5,
        nms_iou: float = 1,
        **kwargs,
    ) -> dict:
        """Predict segmentation.

        Args:
            image: A ndarray of (h,w,c) format. Value of c must be 1-3

        Keyword Args:
            output_type: "label" | "contour" | "bbox"
            min_area: Minimum area of a valid prediction.
            scaling: A image scaling factor. If not 1, the input image will be resized internally before fed
                to the model. The results will be resized back to the scale of the orginal input image.
            score_threshold: Min score needed to be included in the output.
            segmentation_threshold: Threshold value for segmentation.
            nms_iou: IOU threshold value for non-max-supression durint post-processing

        Returns:
            For "label" output:

                - pred_scores: The prediction scores of each instance.
                - pred_label: a 2D image label. 0 is background.

            For "contour" output:

                - pred_scores: The prediction scores of each instance.
                - pred_contours: a list of polygon arrays in x-y format.

            For "bbox" output (ie MaskRCNN):

                - pred_scores: The prediction scores of each instance.
                - pred_bboxes: The bounding-boxes of detected instances in y0x0y1x1 format
                - pred_masks: A 3d array representing segmentation within bboxes
        """
        if not output_type in ("bbox", "label", "contour"):
            raise ValueError(
                f"output_type should be 'patch'|'label'|'contour'. Got {output_type} instead."
            )

        module, params = self.model

        if len(kwargs) > 0:
            apply_fn = _cached_partial(module.apply, **kwargs)
        else:
            apply_fn = module.apply

        ctx = freeze(
            dict(
                apply_fn=apply_fn,
                output_type=output_type,
                segmentation_threshold=segmentation_threshold,
                score_threshold=score_threshold,
                scaling=scaling,
                min_area=min_area,
                remove_out_of_bound=remove_out_of_bound,
                nms_iou=nms_iou,
            )
        )
        preds = _predict(ctx, params, image)
        instance_mask = np.asarray(preds["instance_mask"])

        if output_type == "bbox":
            return dict(
                pred_scores=np.asarray(preds["scores"])[instance_mask],
                pred_bboxes=np.asarray(preds["bboxes"])[instance_mask] / scaling,
                pred_masks=np.asarray(preds["segmentations"])[instance_mask],
            )

        elif output_type == "contour":
            contours = _to_polygons(
                preds,
                segmentation_threshold=segmentation_threshold,
                scaling=scaling,
            )

            return dict(
                pred_scores=np.asarray(preds["scores"])[instance_mask],
                pred_contours=contours,
            )

        else:  # Label
            if "label" in preds:
                mask = np.asarray(preds["instance_mask"])
                label = np.asarray(preds["label"])
                scores = np.asarray(preds["scores"])[mask]

            else:  # compute via polygons
                contours = _to_polygons(
                    preds,
                    segmentation_threshold=segmentation_threshold,
                    scaling=scaling,
                )
                label = _draw_label(contours, image.shape[:2])

            return dict(
                pred_scores=np.asarray(scores),
                pred_label=label,
            )

    def predict_on_large_image(
        self,
        image: ArrayLike,
        gs: int,
        ss: int,
        *,
        scaling: float = 1,
        min_area: int = 0,
        nms_iou: float = 0.25,
        segmentation_threshold: float = 0.5,
        score_threshold: float = 0.5,
        output_type: str = "label",
        min_cells_per_patch: int = 0,
        disable_padding: bool = False,
        **kwargs,
    ) -> dict:
        """Make prediction on very large image by dividing into a grid.

        Direct model prediction on large image may cause out-of-memory error. This
        method divided the large image to smaller grid and then stitch the results
        to form the complete prediction.

        Args:
            image: An image with (H, W, C) format.
            gs: An int value. Grid size of the computation.
            ss: An int value of stepping size. Must be small than gs to produce valid
                results.

        Keyword Args:
            output_type: label" | "contour" | "bbox"
            min_area: Minimum area of a valid prediction.
            scaling: A image scaling factor. If not 1, the input image will be resized internally before fed
                to the model. The results will be resized back to the scale of the orginal input image.
            score_threshold: Min score needed to be included in the output.
            segmentation_threshold: Threshold value for segmentation.
            nms_iou: IOU threshold value for non-max-supression durint post-processing

        Returns:
            For "label" output:

                - pred_scores: The prediction scores of each instance.
                - pred_label: a 2D image label. 0 is background.

            For "contour" output:

                - pred_scores: The prediction scores of each instance.
                - pred_contours: a list of polygon arrays in x-y format.

            For "bbox" output (ie MaskRCNN):

                - pred_scores: The prediction scores of each instance.
                - pred_bboxes: The bounding-boxes of detected instances in y0x0y1x1 format
                - pred_masks: A 3d array representing segmentation within bboxes
        """

        if not output_type in ("bbox", "label", "contour"):
            raise ValueError(
                f"output_type should be 'patch'|'label'|'contour'. Got {output_type} instead."
            )

        module, params = self.model

        if len(kwargs) > 0:
            apply_fn = _cached_partial(module.apply, **kwargs)
        else:
            apply_fn = module.apply

        ctx = freeze(
            dict(
                apply_fn=apply_fn,
                output_type="bbox" if output_type == "bbox" else "grid",
                segmentation_threshold=segmentation_threshold,
                score_threshold=score_threshold,
                scaling=scaling,
                min_area=min_area,
                remove_out_of_bound=True,
                nms_iou=0,
            )
        )

        h, w = image.shape[:2]

        if not disable_padding:
            get_padding = lambda a, gs, ss: (a - 1) // ss * ss + gs - a

            padded_img = np.pad(
                image,
                [[0, get_padding(h, gs, ss)], [0, get_padding(w, gs, ss)], [0, 0]],
            )
        else:
            padded_img = image

        preds = []
        for y0 in range(0, h, ss):
            for x0 in range(0, w, ss):
                logging.info(f"Processing grid {y0}-{x0}")
                pred = _predict(
                    ctx,
                    params,
                    padded_img[y0 : y0 + gs, x0 : x0 + gs],
                )
                pred = _remove_edge_instances(
                    pred,
                    [gs, gs],
                    remove_top=y0 > 0,
                    remove_bottom=(y0 + gs) < h,
                    remove_left=x0 > 0,
                    remove_right=(x0 + gs) < w,
                    threshold=segmentation_threshold,
                )

                logging.info(f"Transfer result...")

                mask = np.asarray(pred["instance_mask"])

                if np.count_nonzero(mask) > min_cells_per_patch:
                    if output_type != "bbox":
                        pred = dict(
                            scores=np.array(pred["scores"])[mask],
                            bboxes=np.array(pred["bboxes"])[mask] + [y0, x0, y0, x0],
                            segmentations=np.array(pred["segmentations"])[mask],
                            y0s=np.array(pred["y0s"])[mask] + y0,
                            x0s=np.array(pred["x0s"])[mask] + x0,
                        )

                        polygons = _to_polygons(
                            pred,
                            segmentation_threshold=segmentation_threshold,
                            scaling=scaling,
                        )

                        valid_polygons = np.asarray([len(p) > 0 for p in polygons])
                        scores = np.where(valid_polygons, pred["scores"], -1)
                        bboxes = np.where(
                            valid_polygons[:, None], pred["bboxes"] / scaling, -1
                        )

                        preds.append(
                            dict(
                                scores=scores,
                                polygons=np.fromiter(polygons, dtype=object),
                                bboxes=bboxes,
                            )
                        )

                    else:
                        bboxes = np.array(pred["bboxes"])[mask] + [y0, x0, y0, x0]
                        bboxes /= scaling
                        preds.append(
                            dict(
                                scores=np.array(pred["scores"])[mask],
                                bboxes=bboxes,
                                segmentations=np.array(pred["segmentations"])[mask],
                            )
                        )

        logging.info(f"Postprocessing...")
        preds = jax.tree_util.tree_map(lambda *x: np.concatenate(x), *preds)

        # sort based on scores
        asort = np.argsort(preds["scores"])[::-1]
        preds = jax.tree_util.tree_map(lambda x: x[asort], preds)

        # nms
        if nms_iou > 0:
            logging.info(f"nms...")
            scores = preds["scores"]
            boxes = preds["bboxes"]
            _, _, selections = sorted_non_max_suppression(
                scores,
                boxes,
                -1,
                threshold=nms_iou,
                min_score=0,
                return_selection=True,
            )

            preds = jax.tree_util.tree_map(lambda x: x[selections], preds)

        if output_type == "bbox": # FIXME unimplemented

            return dict(
                pred_scores=preds["scores"],
                pred_bboxes=preds["bboxes"],
                pred_masks=segs,
            )

        if output_type == "label":
            logging.info(f"Generating label...")

            label = _draw_label(preds["polygons"], image.shape[:2])

            return dict(
                pred_scores=preds["scores"],
                pred_label=label,
            )

        else:  # contours
            return dict(
                pred_scores=preds["scores"],
                pred_contours=preds["polygons"],
            )

    @property
    def module(self) -> nn.Module:
        return self.model[0]

    @property
    def params(self) -> Params:
        return self.model[1]

    @params.setter
    def params(self, new_params):
        self.model = self.module, new_params

    @property
    def detector(self) -> lacss.modules.Detector:
        return self.module.detector

    # FIXME Change the detector without first compile the mode will result
    # in error. This is a hidden gotcha. Need to setup warning to the user.
    # @detector.setter
    # def detector(self, new_detector):
    #     from ..modules import Detector

    #     cur_module = self.module

    #     if isinstance(new_detector, dict):
    #         new_detector = Detector(**new_detector)

    #     if isinstance(new_detector, Detector):
    #         self.model = (
    #             lacss.modules.Lacss(
    #                 self.module.backbone,
    #                 self.module.lpn,
    #                 new_detector,
    #                 self.module.segmentor,
    #             ),
    #             self.model[1],
    #         )
    #     else:
    #         raise ValueError(f"{new_detector} is not a Lacss Detector")

    def save(self, save_path) -> None:
        """Re-save the model by pickling.

        In the form of (module, weights).

        Args:
            save_path: Path to the pkl file
        """
        with open(save_path, "wb") as f:
            pickle.dump(self.model, f)



import struct
import sys
from pathlib import Path
from zipfile import ZipFile

import jax
import jax.numpy as jnp
import numpy as np
import typer

import lacss.deploy.lacss_pb2 as LacssMsg
from lacss.ops import patches_to_label, sorted_non_max_suppression

from .predict import Predictor

app = typer.Typer(pretty_exceptions_enable=False)

def read_input(st = sys.stdin.buffer):
    msg_size = st.read(4)
    msg_size = struct.unpack(">i", msg_size)[0]

    msg = LacssMsg.Input()
    msg.ParseFromString(st.read(msg_size))

    image = msg.image
    np_img = np.frombuffer(image.data, dtype=">f4").astype("float32")
    np_img = np_img.reshape(image.height, image.width, image.channel)
    # np_img = np_img.transpose(2, 1, 0)

    import imageio.v2 as imageio
    imageio.imwrite("tmpin.tif", np_img)

    return np_img, msg.settings

def img_to_msg(msg, img):
    img = np.ascontiguousarray(img, dtype=">i2") # match java format

    msg.height = img.shape[0]
    msg.width = img.shape[1]
    msg.data = img.tobytes()

    assert len(msg.data) == msg.height * msg.width * 2

def write_result(label, score, st = sys.stdout.buffer):

    msg = LacssMsg.Result()
    img_to_msg(msg.score, (score * 1000).astype(np.int16))
    img_to_msg(msg.label, label)

    msg_size_bits = struct.pack(">i", msg.ByteSize())

    st.write(msg_size_bits)
    st.write(msg.SerializeToString())

def write_polygon_result(polygons, scores, st = sys.stdout.buffer):
    msg = LacssMsg.PolygonResult()
    for polygon, score in zip(polygons, scores):
        if len(polygon) > 1:
            polygon_msg = LacssMsg.Polygon()
            polygon_msg.score = score
            for p in polygon:
                point = LacssMsg.Point()
                point.x = p[0]
                point.y = p[1]
                polygon_msg.points.append(point)
            msg.polygons.append(polygon_msg)
    
    msg_size_bits = struct.pack(">i", msg.ByteSize())

    st.write(msg_size_bits)
    st.write(msg.SerializeToString())

@app.command()
def main(modelpath: Path):
    modelpath = str(modelpath)
    if "!" in modelpath: # in a jar
        jarfile, rscpath = modelpath.split("!")
        with ZipFile(jarfile) as jar:
            with jar.open(rscpath) as modelfile:
                import pickle
                modelpath = pickle.load(modelfile)

    model = Predictor(modelpath)
    model.detector.test_max_output = 512
    model.detector.test_min_score = 0

    print(f"lacss_server: loaded model from {modelpath}", file=sys.stderr)
    print(f"lacss_server: default backend is {jax.default_backend()}", file=sys.stderr)
    if (jax.default_backend() == "cpu"):
        print(f"lacss_server: WARNING: No GPU configuration. This might be very slow ...", file=sys.stderr)

    # cnt = 0

    while True:
        img, settings = read_input()
        # img = img - img.min()
        # img = img / img.max()
        img -= img.mean()
        img = img / img.std()

        print(f"received image {img.shape}", file=sys.stderr)

        preds = model.predict(
            img, 
            min_area=settings.min_cell_area,
            remove_out_of_bound=settings.remove_out_of_bound,
            scaling=settings.scaling,
            nms_iou=settings.nms_iou,
            score_threshold=settings.detection_threshold,
            segmentation_threshold=settings.segmentation_threshold,
            output_type="contour" if settings.return_polygon else "label"
        )

        if settings.return_polygon:

            write_polygon_result(
                preds["pred_contours"],
                preds["pred_scores"],
            )

        else:

            label = preds["pred_label"].astype("int")
            scores = preds["pred_scores"]

            assert label.max() == len(scores)

            assert len(label.shape) == 2

            # score image
            score_img = scores[label - 1]
            score_img *= (label > 0)

            assert score_img.shape == label.shape

            write_result(label, score_img)

        # imageio.imwrite(f"p_{cnt}.tif", np.asarray(label))
        # cnt+=1

if __name__ == "__main__":
    app()
